import json
import os
import pickle
import sys

from datasets import DownloadManager

if __name__ == "__main__":
    output_file = sys.argv[1]

    PATH_TO_REPOSITORY = "https://nlp.informatik.hu-berlin.de/resources/datasets/zelda/zelda_full.zip"
    data_dir = os.path.join(DownloadManager().download_and_extract(PATH_TO_REPOSITORY), "zelda")
    counter = pickle.load(open(os.path.join(data_dir, 'other', 'zelda_mention_entities_counter.pickle'), 'rb'))
    with open(output_file, 'w') as f:
        json.dump(counter, f)



import logging
import math
import os
from dataclasses import dataclass
from typing import Any, Optional

import torch
from datasets import Dataset
from tqdm.auto import tqdm
from transformers import (
    AutoTokenizer,
    EvalPrediction,
    set_seed,
)

from entity_linkings.data_utils import EntityDictionary
from entity_linkings.trainer import EntityLinkingTrainer, TrainingArguments
from entity_linkings.utils import BaseSystemOutput, calculate_recall_mrr

from ..base import RetrieverBase
from ..collator import CollatorForRetrieval
from .encoder import DualBERTModel
from .indexer import DenseRetriever
from .preprocessor import DualEncoderPreprocessor

logger = logging.getLogger(__name__)


def compute_metrics(p: EvalPrediction) -> dict[str, Any]:
    scores = p.predictions
    preds = scores.argmax(axis=1).ravel()
    labels = p.label_ids.ravel()
    mask = labels != -100
    preds = preds[mask]
    labels = labels[mask]

    num_corrects = (preds == labels).sum().item()
    num_golds = mask.sum().item()
    recall = num_corrects / num_golds if num_golds > 0 else float("nan")
    return {"recall": recall}


class DUALENCODER(RetrieverBase):
    '''
    Span-level Entity Retrieval with Dual-BERT Encoder
    '''

    @dataclass
    class Config(RetrieverBase.Config):
        '''Retrieval configuration
        Args:
            - model_name_or_path (str): Pre-trained model name or path for mention and entity encoders
            - ent_start_token (str): Special token to indicate the start of entity mention
            - ent_end_token (str): Special token to indicate the end of entity mention
            - entity_token (str): Special token to indicate the entity
            - nil_token (str): Special token for NIL entity
            - max_context_length (int): Maximum sequence length for mention context
            - max_candidate_length (int): Maximum sequence length for entity context
            - pooling (str): Pooling method for obtaining fixed-size representations
            - metric (str): Similarity metric for retrieval ('inner_product', 'cosine', 'euclidean')
            - num_hard_negatives (int): Number of hard negatives to sample during training
            - temperature (float): Temperature for scaling similarity scores
        '''
        model_name_or_path: Optional[str] = "google-bert/bert-base-uncased"
        ent_start_token: str = "[START_ENT]"
        ent_end_token: str = "[END_ENT]"
        entity_token: str = "[ENT]"
        nil_token: str = "[NIL]"
        max_context_length: int = 128
        max_candidate_length: int = 50
        index_batch_size: int = 128
        pooling: str = 'first'
        distance: str = 'inner_product'
        temperature: float = 1.0
        context_window_chars: int = 500
        n_hubs: int = 10
        use_hnsw: bool = False
        fp16: bool = False

    def __init__(self, dictionary: EntityDictionary, config: Optional[Config] = None, index_path: Optional[str] = None) -> None:
        super().__init__(dictionary, config)
        self.tokenizer = AutoTokenizer.from_pretrained(self.config.model_name_or_path)
        if os.path.exists(self.config.model_name_or_path):
            logger.info(f"Loading model from {self.config.model_name_or_path}")
            self.encoder = DualBERTModel.from_pretrained(self.config.model_name_or_path)
        else:
            special_tokens = [self.config.ent_start_token, self.config.ent_end_token, self.config.entity_token, self.config.nil_token]
            self.tokenizer.add_special_tokens({"additional_special_tokens": special_tokens})
            self.encoder = DualBERTModel(
                model_name_or_path=self.config.model_name_or_path,
                pooling=self.config.pooling,
                distance=self.config.distance,
                temperature=self.config.temperature,
            )
            self.encoder.resize_token_embeddings(len(self.tokenizer))
        self.preprocessor = DualEncoderPreprocessor(
            self.tokenizer, self.config.ent_start_token, self.config.ent_end_token,
            self.config.entity_token, self.config.max_context_length,
            self.config.max_candidate_length, self.config.context_window_chars
        )
        self.dictionary = self.preprocessor.dictionary_preprocess(self.dictionary)
        if index_path is not None:
            self.retriever = self.create_retriever(index_path=index_path)

    def create_retriever(self, index_path: str) -> DenseRetriever:
        retriever = DenseRetriever(
            dictionary=self.dictionary,
            model=self.encoder,
            tokenizer=self.tokenizer,
            batch_size=self.config.index_batch_size,
            metric=self.config.distance,
            use_hnsw=self.config.use_hnsw,
            n_hubs=self.config.n_hubs,
            fp16=self.config.fp16
        )
        retriever.build_index(index_path=index_path)
        return retriever

    def train(
            self,
            train_dataset: Dataset,
            eval_dataset: Optional[Dataset] = None,
            num_hard_negatives: int = 0,
            training_args: Optional[TrainingArguments] = None
        ) -> dict[str, float]:
        if training_args is None:
            training_args = TrainingArguments()
        set_seed(training_args.seed)

        if num_hard_negatives > 0:
            train_candidates = self.retrieve_candidates(
                train_dataset,
                only_negative=True,
                batch_size=training_args.per_device_eval_batch_size,
                top_k=num_hard_negatives,
            )
            if eval_dataset is not None:
                eval_candidates = self.retrieve_candidates(
                    eval_dataset,
                    only_negative=True,
                    batch_size=training_args.per_device_eval_batch_size,
                    top_k=num_hard_negatives,
                )

        processed_train_dataset = self.preprocessor.dataset_preprocess(train_dataset, train_candidates if num_hard_negatives > 0 else None)
        if eval_dataset is not None:
            processed_eval_dataset = self.preprocessor.dataset_preprocess(eval_dataset, eval_candidates if num_hard_negatives > 0 else None)

        trainer = EntityLinkingTrainer(
            model=self.encoder,
            args=training_args,
            data_collator=CollatorForRetrieval(
                self.tokenizer,
                dictionary=self.dictionary,
                num_hard_negatives=num_hard_negatives
            ),
            train_dataset=processed_train_dataset,
            eval_dataset=processed_eval_dataset if eval_dataset is not None else None,
            compute_metrics=compute_metrics
        )
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint

        results = trainer.train(resume_from_checkpoint=checkpoint)
        trainer.log_metrics("train", results.metrics)
        if training_args.output_dir is not None:
            self.encoder.save_pretrained(training_args.output_dir)
            self.tokenizer.save_pretrained(training_args.output_dir)
            trainer.save_state()
            trainer.save_metrics("train", results.metrics)
        return results

    @torch.no_grad()
    def evaluate(self, dataset: Dataset, batch_size: int = 32, **args: int) -> dict[str, float]:
        self.encoder.eval()
        queries, labels = [], []
        for text, entities in zip(dataset["text"], dataset["entities"]):
            for ent in entities:
                marked_text = self.preprocessor._process_context(text, ent["start"], ent["end"])
                queries.append(marked_text)
                labels.append(ent["label"])

        pbar = tqdm(total=(math.ceil(len(queries)//batch_size)), desc='Evaluate')
        predictions = []
        for i in range(0, len(queries), batch_size):
            pbar.update()
            _, batch_indices = self.retriever.search_knn(queries[i: i + batch_size], top_k=100)
            batch_labels = labels[i: i + batch_size]
            for j, indices in enumerate(batch_indices):
                preds = [{"id": self.dictionary(inds)["id"]} for inds in indices]
                predictions.append({"gold": batch_labels[j], "predict": preds})
        pbar.close()
        metric = calculate_recall_mrr(predictions)
        return metric

    @torch.no_grad()
    def predict(self, sentence: str, spans: Optional[list[tuple[int, int]]] = None, top_k: int = 5) -> list[list[BaseSystemOutput]]:
        if not spans:
            raise ValueError("Spans must be provided for SpanEntityRetrieval prediction.")

        self.encoder.eval()
        inputs = [self.preprocessor._process_context(sentence, b, e) for b, e in spans]
        similarities, indices = self.retriever.search_knn(inputs, top_k=top_k)

        all_result = []
        for i, (b, e) in enumerate(spans):
            result = []
            for _, ind in enumerate(indices[i]):
                entry = self.dictionary(ind)
                result.append(BaseSystemOutput(query=sentence[b:e], start=b, end=e, id=entry['id']))
            all_result.append(result)
        return all_result

    @torch.no_grad()
    def retrieve_candidates(self, dataset: Dataset, top_k: int = 5, only_negative: bool = False, batch_size: int = 32, **args: int) -> list[list[str]]:
        self.encoder.eval()
        queries, labels = [], []
        for text, entities in zip(dataset["text"], dataset["entities"]):
            for ent in entities:
                query = self.preprocessor._process_context(text, ent["start"], ent["end"])
                queries.append(query)
                labels.append(ent["label"])

        candidates = []
        pbar  = tqdm(total=len(queries), desc='Retrieve candidates')
        for i in range(0, len(queries), batch_size):
            pbar.update(min(batch_size, len(queries[i])))
            _, batch_indices = self.retriever.search_knn(
                queries[i: i + batch_size],
                top_k=top_k,
                ignore_ids=labels[i: i + batch_size] if only_negative else None
            )
            candidates.extend(batch_indices)
        pbar.close()
        return candidates



import tempfile
from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers.trainer_utils import TrainOutput

import assets as test_data
from entity_linkings import get_rerankers, get_retrievers, load_dictionary
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import BaseSystemOutput

from .fevry import FEVRY

MODELS = ["google-bert/bert-base-uncased"]
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)


@pytest.mark.reranker_fevry
class TestFEVRY:
    def test___init__(self) -> None:
        model_cls = get_rerankers("fevry")
        model = model_cls(retriever=retriever, config=FEVRY.Config(model_name_or_path=MODELS[0]))
        assert isinstance(model, FEVRY)
        assert model.config.model_name_or_path == MODELS[0]
        assert "[NIL]" in model.tokenizer.all_special_tokens
        assert model.config.max_context_length == 128
        assert model.config.context_window_chars == 500

    def test_train(self) -> None:
        model = FEVRY(retriever=retriever, config=FEVRY.Config(model_name_or_path=MODELS[0]))
        with tempfile.TemporaryDirectory() as tmpdir:
            result = model.train(
                train_dataset=dataset,
                eval_dataset=dataset,
                num_candidates=3,
                training_args=TrainingArguments(
                    output_dir=tmpdir,
                    num_train_epochs=1,
                    per_device_train_batch_size=2,
                    per_device_eval_batch_size=2,
                    logging_strategy="no",
                    save_strategy="no",
                    eval_strategy="no",
                    remove_unused_columns=False,
                    eval_on_start=True
                )
            )
            assert isinstance(result, TrainOutput)
            assert hasattr(result, 'metrics')

    def test_evaluate(self) -> None:
        model = FEVRY(retriever=retriever, config=FEVRY.Config(model_name_or_path=MODELS[0]))
        metrics = model.evaluate(dataset, num_candidates=3, batch_size=2)
        assert 'top1_accuracy' in metrics

    def test_predict(self) -> None:
        sentence = "Apple is looking at buying U.K. startup for $1 billion"
        spans = [(0,5), (27,30), (44,52)]
        model = FEVRY(retriever=retriever, config=FEVRY.Config(model_name_or_path=MODELS[0]))
        predictions = model.predict(sentence, spans=spans, num_candidates=3)
        assert isinstance(predictions, list)
        assert len(predictions) == len(spans)
        for preds in predictions:
            assert isinstance(preds, BaseSystemOutput)
            assert preds.id is not None and preds.start is not None and preds.end is not None



import tempfile
from importlib.resources import files

import numpy as np
import pytest
from datasets import load_dataset

import assets as test_data
from entity_linkings import load_dictionary

from .indexer import MentionPriorIndexer

dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
mention_counter_path = str(files(test_data).joinpath("mention_counter_toy.json"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']


@pytest.fixture(scope='module')
def mention_prior_indexer() -> MentionPriorIndexer:
    indexer = MentionPriorIndexer(dictionary=dictionary, mention_counter_path=mention_counter_path)
    with tempfile.TemporaryDirectory() as tmpdir:
        indexer.build_index(tmpdir)
    return indexer

class TestMentionPriorIndexer:
    def test_build_index(self) -> None:
        indexer = MentionPriorIndexer(dictionary=dictionary, mention_counter_path=mention_counter_path)
        with tempfile.TemporaryDirectory() as tmpdir:
            indexer.build_index(tmpdir)
            assert len(indexer) == len(dictionary)
            assert len(list(indexer.meta_ids_to_keys.keys())) == len(dictionary)

    @pytest.mark.parametrize("top_k", [0, 1, 2])
    def test_search_knn(self, top_k: int) -> None:
        queries = ["Microsoft", "Meta", "nahanaha"]
        if top_k <= 0:
            indexer = MentionPriorIndexer(dictionary=dictionary, mention_counter_path=mention_counter_path)
            with tempfile.TemporaryDirectory() as tmpdir:
                indexer.build_index(tmpdir)
                with pytest.raises(RuntimeError) as re:
                    indexer.search_knn(queries, top_k)
                assert isinstance(re.value, RuntimeError)
                assert str(re.value) == "K is zero or under zero."
        else:
            indexer = MentionPriorIndexer(dictionary=dictionary, mention_counter_path=mention_counter_path)
            with tempfile.TemporaryDirectory() as tmpdir:
                indexer.build_index(tmpdir)
                distances, indices = indexer.search_knn(queries, top_k)
                assert isinstance(distances, np.ndarray) and isinstance(indices, list)
                if top_k > len(dictionary):
                    assert distances.shape[0] == len(indices) == 3
                    assert distances.shape[1] == len(indices[0]) == len(dictionary)
                else:
                    assert distances.shape[0] == len(indices) == 3
                    assert distances.shape[1] == len(indices[0]) == top_k

    @pytest.mark.parametrize("top_k", [1])
    def test_search_knn_negatives(self, mention_prior_indexer: MentionPriorIndexer, top_k: int) -> None:
        for example in dataset:
            if not example["entities"]:
                continue
            ignore_ids = [entity["label"] for entity in example["entities"]]
            queries = [example["text"][ent["start"]: ent["end"]] for ent in example["entities"]]
            _, indices = mention_prior_indexer.search_knn(queries, top_k, ignore_ids=ignore_ids)
            for i, inds in enumerate(indices):
                assert len(inds) == top_k
                for ind in inds:
                    assert ind not in ignore_ids[i]

    def test_len(self, mention_prior_indexer: MentionPriorIndexer) -> None:
        assert len(mention_prior_indexer) == len(dictionary)



import json
import os
from logging import getLogger
from typing import Literal, Optional, Union

import bm25s
import numpy as np
from bm25s.hf import BM25HF
from bm25s.tokenization import Tokenized

from entity_linkings.data_utils import EntityDictionary

from ..base import IndexerBase

logger = getLogger(__name__)


class BM25Indexer(IndexerBase):
    '''
    BM25 retriever using bm25s (https://github.com/xhluca/bm25s)
    '''

    def __init__(
            self,
            dictionary: EntityDictionary,
            language: str = "en",
            n_threads: int = -1,
            subword_tokenizer: Optional[str] = None,
            query_type_for_candidate: Literal['mention', 'description'] = 'mention'
        ) -> None:
        super().__init__(dictionary)
        if subword_tokenizer is not None:
            from transformers import AutoTokenizer
            self.tokenizer = AutoTokenizer.from_pretrained(subword_tokenizer)
            self.tokenize_func = self.subword_tokenize
        else:
            if language == "en":
                self.tokenize_func = self.whitespace_tokenize
            else:
                raise NotImplementedError(f"Language {language} is not supported yet.")
        self.n_threads = n_threads
        self.query_type_for_candidate = query_type_for_candidate

    def whitespace_tokenize(self, texts: Union[str, list[str]]) -> Tokenized:
        def _tokenize(text: list[str]) -> list[str]:
            return [token for token in text]
        corpus_token = bm25s.tokenize(texts, token_pattern=r"(?u)\b[\w#]+\b", stopwords=[], stemmer=_tokenize)
        return corpus_token

    def subword_tokenize(self, texts: Union[str, list[str]]) -> Tokenized:
        corpus_ids = []
        tokens_to_idx: dict[str, int] = {}
        for text in texts:
            tokens = self.tokenizer.tokenize(text, add_special_tokens=False)
            doc_ids = []
            for token in tokens:
                if token not in tokens_to_idx:
                    tokens_to_idx[token] = len(tokens_to_idx)
                token_id = tokens_to_idx[token]
                doc_ids.append(token_id)
            corpus_ids.append(doc_ids)
        return Tokenized(ids=corpus_ids, vocab=tokens_to_idx)

    def _initialize(self) -> None:
        self.index = BM25HF()
        self.meta_ids_to_keys: dict[int, str] = {}

    def build_index(self, index_path: str) -> None:
        if os.path.exists(os.path.join(index_path, "meta_bm25.json")):
            self.load(index_path)
        else:
            self._initialize()
            descriptions = []
            for i, entity in enumerate(self.dictionary):
                self.meta_ids_to_keys[i] = entity["id"]
                if self.query_type_for_candidate == 'mention':
                    descriptions.append(entity['name'])
                else:
                    descriptions.append(f"{entity['name']}: {entity['description']}")
            corpus_tokens = self.tokenize_func(descriptions)
            self.index.index(corpus_tokens)

    def search_knn(self, query: str|list[str], top_k: int, ignore_ids: Optional[list[str]|list[list[str]]] = None) -> tuple[np.ndarray, list[list[str]]]:
        if top_k <= 0:
            raise RuntimeError("K is zero or under zero.")
        if top_k > len(self.meta_ids_to_keys):
            top_k = len(self.meta_ids_to_keys)
            logger.warning(f"K is over size of dictionary. K is modified to size of dictionary to {len(self.meta_ids_to_keys)}")

        additional_top_k = 0
        if ignore_ids is not None:
            if isinstance(ignore_ids[0], list):
                additional_top_k = max([len(ids) for ids in ignore_ids])
            else:
                additional_top_k = len(ignore_ids)
        query_tokens = self.tokenize_func(query)
        results, scores = self.index.retrieve(
            query_tokens, k=top_k+additional_top_k, n_threads=self.n_threads
        )

        indices_keys = []
        for i in range(results.shape[0]):
            if not ignore_ids:
                indices_keys.append([self.meta_ids_to_keys[j] for j in results[i].tolist()])
                continue
            candidate_ids = []
            for j in results[i].tolist():
                key = self.meta_ids_to_keys[j]
                if key not in ignore_ids[i]:
                    candidate_ids.append(key)
            indices_keys.append(candidate_ids[:top_k])
        return scores, indices_keys

    def save_index(self, index_path: str, ensure_ascii: bool = False) -> None:
        logger.info("Serializing index to %s", index_path)
        self.index.save(index_path)
        meta_file = os.path.join(index_path, "meta_bm25.json")
        json.dump(self.meta_ids_to_keys, open(meta_file, 'w'), ensure_ascii=ensure_ascii)

    def load(self, index_path: str) -> None:
        self._initialize()
        logger.info("Deserializing index from %s", index_path)
        self.index = BM25HF.load(index_path, load_corpus=True)
        meta_file = os.path.join(index_path, "meta_bm25.json")
        self.meta_ids_to_keys.update({int(k): v for k, v in json.load(open(meta_file)).items()})
        logger.info(
            "Loaded index of type %s and size %d", type(self.index), len(self.meta_ids_to_keys)
        )

    def __len__(self) -> int:
        return len(self.meta_ids_to_keys)



import json
import os
from argparse import ArgumentParser
from typing import Any

from tqdm.auto import tqdm

from appendix.data_convert.utils import WikiMapper


def get_documents(input_path: str, id_col: int, text_col: int) -> dict[str, str]:
    documents: dict[str, str] = {}
    post_id = "-1"
    with open(input_path, 'r', encoding='utf-8') as f:
        for line in f:
            cols = line.split('\t')
            if len(cols) == 1:
                documents[post_id] += cols[0]
                continue
            doc_id = cols[id_col]
            documents[doc_id] = cols[text_col]
            post_id = doc_id
    return documents


def get_annotation(input_path: str) -> dict[str, Any]:
    annotations: dict[str, Any] = {}

    with open(input_path, 'r', encoding='utf-8') as f:
        lines = f.readlines()

    for line in lines:
        if line.strip():
            cols = line.strip().split('\t')
            doc_id = cols[0]
            mention = cols[2]
            wikiname = cols[3]
            start = int(cols[4])
            end = int(cols[5])
            if doc_id not in annotations:
                annotations[doc_id] = []
            annotations[doc_id].append({
                "mention": mention,
                "start": start,
                "end": end,
                "wikiname": wikiname,
            })
    return annotations


def _convert(
        raw_examples: dict[str, str],
        annotations: dict[str, Any],
        finder: WikiMapper,
        subset: str
    ) -> list[dict[str, Any]]:
    examples = []
    pbar = tqdm(total=len(raw_examples))
    for doc_id, content in raw_examples.items():
        pbar.update()
        entities = []
        content = content.strip()
        for entity in annotations.get(doc_id, []):
            wikiname = entity["wikiname"]
            res = finder.find_by_title(wikiname)
            page_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
            assert content[entity["start"]:entity["end"]] == entity["mention"], f"{content[entity['start']:entity['end']]} != {entity['mention']}"

            entities.append({
                "start": entity["start"],
                "end": entity["end"],
                "label": [str(page_id)],
            })

        examples.append({
            "subset": subset,
            "id": doc_id,
            "text": content,
            "entities": entities,
        })
    pbar.close()
    return examples


def convert_posts(input_dir: str, output_dir: str, finder: WikiMapper) -> None:
    raw_examples = get_documents(os.path.join(input_dir, "posts.tsv"), id_col=0, text_col=2)
    annotations = get_annotation(os.path.join(input_dir, "gold_post_annotations.tsv"))
    posts_examples = _convert(raw_examples, annotations, finder, subset="posts")
    with open(os.path.join(output_dir, "posts.jsonl"), 'w', encoding='utf-8') as f:
        for item in posts_examples:
            f.write(json.dumps(item, ensure_ascii=False) + '\n')


def convert_comments(input_dir: str, output_dir: str, finder: WikiMapper) -> None:
    raw_examples = get_documents(os.path.join(input_dir, "comments.tsv"), id_col=0, text_col=4)
    annotations = get_annotation(os.path.join(input_dir, "gold_comment_annotations.tsv"))
    comments_examples = _convert(raw_examples, annotations, finder, subset="comments")
    with open(os.path.join(output_dir, "comments.jsonl"), 'w', encoding='utf-8') as f:
        for item in comments_examples:
            f.write(json.dumps(item, ensure_ascii=False) + '\n')


def main(input_dir: str, output_dir: str, finder: WikiMapper) -> None:
    output_dir_data = os.path.join(output_dir, 'reddit')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    convert_posts(input_dir, output_dir_data, finder)
    convert_comments(input_dir, output_dir_data, finder)


if __name__ == "__main__":
    parser = ArgumentParser(description="Reddit Entity Disambiguation Script")
    parser.add_argument("--input_dir", "-i", type=str, required=True, help="Path to the input text file")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(args.input_dir, args.output_dir, finder)
    finder.close()



import logging
from dataclasses import dataclass
from typing import Any, Optional, Union

from transformers import BatchEncoding, PreTrainedTokenizer
from transformers.data.data_collator import pad_without_fast_tokenizer_warning
from transformers.utils import PaddingStrategy

logger = logging.getLogger(__name__)


@dataclass
class CollatorBase:
    tokenizer: PreTrainedTokenizer
    padding: Union[bool, str, PaddingStrategy] = True
    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    return_tensors: str = "pt"

    def __call__(self, features: list[Union[dict[str, Any], BatchEncoding]]) -> BatchEncoding:
        features = [f.copy() for f in features]
        encodings: list[dict[str, Any]] = []
        labels = []
        for f in features:
            encoding = f.get('encoding', None)
            if encoding is None:
                input_ids = f.get('input_ids', None)
                attention_mask = f.get('attention_mask', None)
                token_type_ids = f.get('token_type_ids', None)
                if input_ids is None or attention_mask is None:
                    raise ValueError("input_ids and attention_mask fields are required in features")
                encoding = {'input_ids': input_ids, 'attention_mask': attention_mask}
                if token_type_ids is not None:
                    encoding['token_type_ids'] = token_type_ids
            encodings.append(encoding)
            label = f.get('labels', None)
            if label is not None:
                labels.append(label)
        batch = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            encodings,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors=self.return_tensors,
        )
        if labels:
            batch['labels'] = labels
        return batch



import itertools
import json
import os
from argparse import ArgumentParser

from tqdm import tqdm

from appendix.data_convert.utils import WikiMapper, read_conll


def main(finder: WikiMapper, input_file: str, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'derczynski')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    with open(os.path.join(output_dir_data, "test.jsonl"), 'w', encoding='utf-8') as f:
        uid = map(str, itertools.count(start=0, step=1))
        for _, sentences in read_conll(input_file, delimiter='\t', tag_column=2, link_column=1):
            pbar = tqdm(total=len(sentences), desc="Processing sentences")
            for sentence in sentences:
                pbar.update()
                entities = []
                for ent in sentence['entities']:
                    if ent['title'][0] == 'NIL':
                        entities.append({"start": ent['start'], "end": ent['end'], "label": ["-1"]})
                        continue
                    title = ent['title'][0].split('/')[-1]
                    res = finder.find_by_title(title)
                    page_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
                    entities.append({
                        "start": ent['start'],
                        "end": ent['end'],
                        "label": [str(page_id)],
                    })
                example = {
                    "subset": "derczynski",
                    "id": next(uid),
                    "text": sentence['text'],
                    "entities": entities,
                }
                f.write(json.dumps(example, ensure_ascii=False) + "\n")
            pbar.close()


if __name__ == "__main__":
    parser = ArgumentParser(description="IPM-NEL2023 Script")
    parser.add_argument("--input_file", "-i", type=str, default="ipm_nel.conll", help="Path to the input text file")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, args.input_file, output_dir=args.output_dir)
    finder.close()




from typing import Literal

import numpy as np
import torch
from transformers import EvalPrediction


def compute_metrics(p: EvalPrediction) -> dict[str, float]:
    start_logits, end_logits = p.predictions
    start_positions, end_positions = p.label_ids # (2, batch_size, n_samples)
    start_positions = start_positions.ravel() # (n_samples, )
    end_positions = end_positions.ravel() # (n_samples, )

    start_predictions = start_logits.argmax(axis=-1).ravel() # (n_samples, )
    end_predictions = end_logits.argmax(axis=-1).ravel() # (n_samples, )
    correct_full_predictions = np.logical_and(
        np.equal(start_predictions, start_positions),
        np.equal(end_predictions, end_positions),
    )
    accuracy = correct_full_predictions.mean().item()
    return {"accuracy": accuracy}


def calculate_joint_probabilities(
    mode: Literal["max-prod", "max-end", "max-start", "max"],
    possible_indices: list[tuple[int, int]],
    classification_probabilities: torch.Tensor,
) -> list[float]:
    def max_prod(x: tuple[int, int]) -> torch.Tensor:
        prob =  classification_probabilities[0][x[0]] * classification_probabilities[1][x[1]]
        return prob

    def max_start(x: tuple[int, int]) -> torch.Tensor:
        prob = classification_probabilities[0][x[0]]
        return prob

    def max_end(x: tuple[int, int]) -> torch.Tensor:
        prob = classification_probabilities[1][x[1]]
        return prob

    def max_func(x: tuple[int, int]) -> torch.Tensor:
        prob =  max(
            classification_probabilities[0][x[0]],
            classification_probabilities[1][x[1]],
        )
        return prob

    if mode == "max-prod":
        selector = max_prod
    elif mode == "max-end":
        selector = max_end
    elif mode == "max-start":
        selector = max_start
    elif mode == "max":
        selector = max_func
    else:
        raise NotImplementedError

    return [selector(x).item() for x in possible_indices]


def select_indices(
    mode: Literal["max-prod", "max-end", "max-start", "max"],
    possible_indices: list[tuple[int, int]],
    classification_probabilities: torch.Tensor,
) -> tuple[int, int]:

    prob_indices = calculate_joint_probabilities(
        mode=mode,
        possible_indices=possible_indices,
        classification_probabilities=classification_probabilities,
    )
    max_indice = prob_indices.index(max(prob_indices))
    return possible_indices[max_indice]



from importlib.resources import files

import pytest
from datasets import load_dataset

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary
from entity_linkings.utils import BaseSystemOutput

from .chatel import CHATEL
from .utils import process_multi_choice_prompt

MODELS = ["gpt-3.5-turbo"]
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)
candidates_ids = retriever.retrieve_candidates(dataset, top_k=3)


@pytest.mark.reranker_chatel
class TestChatEL:
    @pytest.fixture
    def model(self) -> CHATEL:
        return CHATEL(
            retriever=retriever,
            config=CHATEL.Config(
                model_name_or_path=MODELS[0],
                token="dummy-key"
            ),
        )

    def test__init__(self, model: CHATEL) -> None:
        assert hasattr(model, "config") and hasattr(model, "model") and hasattr(model, "dictionary")
        assert model.retriever == retriever
        assert model.config.model_name_or_path == MODELS[0]
        assert model.config.max_generation_length == 100

    def test_first_step_prompt(self, model: CHATEL) -> None:
        sentence = "This is a mention in the text about entity A."
        mention = "entity A"
        prompt = model.first_step_prompt(sentence, sentence.find(mention), sentence.find(mention) + len(mention))
        assert prompt == "This is a mention in the text about entity A. \n What does entity A in this sentence referring to?"

    def test_second_step_prompt(self, model: CHATEL) -> None:
        first_result = "Entity A refers to ..."
        mention = "entity A"
        candidates = [model.dictionary(c) for c in candidates_ids[0]]
        multi_choice_prompt = ""
        for i, cand in enumerate(candidates):
            description = cand['name'] + ' ' + cand['description'][:200] if cand['description'] else cand['name']
            multi_choice_prompt += f"({i+1}). {description}\n"

        second_prompt = model.second_step_prompt(first_result, mention, candidates_ids[0])
        assert second_prompt == f"Entity A refers to ...\n\nWhich of the following does {mention} referes to?\n\n{multi_choice_prompt}"

    def test_predict(self) -> None:
        sentence = "Apple is looking at buying U.K. startup for $1 billion"
        spans = [(0,5), (27,30), (44,52)]
        candidates = retriever.predict(sentence, spans, top_k=3)
        second_results = [
            "(1)",
            "(2)",
            "(3)"
        ]

        all_result = []
        for i, result in enumerate(second_results):
            assert isinstance(candidates[i], list) and len(candidates[i]) == 3
            assert isinstance(candidates[i][0], BaseSystemOutput)
            pred = process_multi_choice_prompt(result, [c.id for c in candidates[i]])
            all_result.append([candidates[i][pred]])
        assert len(all_result) == 3
        assert len(all_result[0]) == 1
        assert all_result[0][0].id == "000012"



import json
import os
from argparse import ArgumentParser
from typing import Any
from xml.etree import ElementTree

from datasets import DownloadManager

from appendix.data_convert.utils import WikiMapper

_URL = "https://github.com/dice-group/gerbil/releases/download/v1.2.6/gerbil_data.zip"



def get_documents(data_dir: str) -> dict[str, str]:
    file_names = os.listdir(data_dir)
    documents = {}
    for file_name in file_names:
        doc_id = file_name.split(".")[0]
        with open(os.path.join(data_dir, file_name), 'r', encoding='utf-8') as f:
            documents[doc_id] = f.read()
    return documents


def parse_annotation(file_path: str, finder: WikiMapper) -> list[dict[str, Any]]:
    entities = []
    with open(file_path, 'r', encoding='utf-8') as f:
        g = ElementTree.parse(f)
        root = g.getroot()
        for anno in root.findall('ReferenceInstance'):
            start = anno.findtext('Offset')
            length = anno.findtext('Length')
            wikilink = anno.findtext('ChosenAnnotation')
            surface_form = anno.findtext('SurfaceForm')
            if start is None or length is None or wikilink is None or surface_form is None:
                raise ValueError("Missing required annotation fields.")

            res = finder.find_by_title(wikilink.split('/')[-1].strip())
            page_id = str(res["page_id"]) if res is not None and res["page_id"] is not None else "-1"

            entities.append({
                "start": int(start),
                "end": int(start) + int(length),
                "label": [page_id],
            })
    return entities


def main(finder: WikiMapper, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'msnbc')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    data_dir = os.path.join(DownloadManager().download_and_extract(_URL), "gerbil_data", "datasets", "MSNBC")
    raw_documents = get_documents(os.path.join(data_dir, "RawTextsSimpleChars_utf8"))

    with open(os.path.join(output_dir_data, "test.jsonl"), 'w', encoding='utf-8') as f:
        for kid, content in raw_documents.items():
            entities = parse_annotation(os.path.join(data_dir, "Problems", f"{kid}.txt"), finder)
            example = {
                "subset": "msnbc",
                "id": kid,
                "text": content,
                "entities": entities
            }
            f.write(f"{json.dumps(example, ensure_ascii=False)}\n")


if __name__ == "__main__":
    parser = ArgumentParser(description="MSNBC Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, output_dir=args.output_dir)
    finder.close()



import json
import os
from typing import Optional

import torch
import torch.nn as nn

from entity_linkings.utils import load_model


class SpanClassifier(nn.Module):

    # to suppress an AttributeError when training
    _keys_to_ignore_on_save = None
    def __init__(self, model_name_or_path: str, num_entities: int, projection_dim: int = 256) -> None:
        super().__init__()
        self.encoder = load_model(model_name_or_path)

        self.projection = torch.nn.Linear(self.encoder.config.hidden_size * 2, projection_dim)
        classifier_dropout = (
            self.encoder.config.classifier_dropout
            if hasattr(self.encoder.config, "classifier_dropout") and self.encoder.config.classifier_dropout is not None
            else self.encoder.config.hidden_dropout_prob
        )
        self.dropout = nn.Dropout(classifier_dropout)

        self.entity_embeddings = nn.Embedding(num_entities, projection_dim)
        self.entity_bias = nn.Embedding(num_entities, 1)
        self.config = {"model_name_or_path": model_name_or_path, "num_entities": num_entities, "projection_dim": projection_dim}

    def resize_token_embeddings(self, new_num_tokens: int) -> None:
        self.encoder.resize_token_embeddings(new_num_tokens)

    def encode(
            self,
            input_ids: torch.Tensor,
            attention_mask: torch.Tensor,
            start_positions: torch.Tensor,
            end_positions: torch.Tensor,
            token_type_ids: Optional[torch.Tensor] = None,
        ) -> torch.Tensor:
        bs, _ = input_ids.size()
        if token_type_ids is not None:
            encoder_outputs = self.encoder(input_ids=input_ids, attention_mask=attention_mask, token_type_ids=token_type_ids)
        else:
            encoder_outputs = self.encoder(input_ids=input_ids, attention_mask=attention_mask)
        last_hidden_state = encoder_outputs.last_hidden_state # (bs, length, hidden_size)

        span_start_embeds = last_hidden_state[torch.arange(bs), start_positions, :]
        span_end_embeds = last_hidden_state[torch.arange(bs), end_positions, :]
        span_embeds = torch.cat([span_start_embeds, span_end_embeds], dim=-1)

        return span_embeds # (bs, hidden_size * 2)

    def forward(
            self,
            input_ids: torch.Tensor,
            attention_mask: torch.Tensor,
            token_type_ids: Optional[torch.Tensor] = None,
            start_positions: Optional[torch.Tensor] = None,
            end_positions: Optional[torch.Tensor] = None,
            candidates_ids: Optional[torch.Tensor] = None,
            labels: Optional[torch.Tensor] = None,
        ) -> tuple[Optional[torch.Tensor], torch.Tensor]:

        span_embeds = self.encode(
            input_ids=input_ids,
            attention_mask=attention_mask,
            start_positions=start_positions,
            end_positions=end_positions,
            token_type_ids=token_type_ids
        )  # (bs, hidden_size * 2)

        span_embeds = self.projection(span_embeds) # (bs, 256)
        span_embeds = self.dropout(span_embeds) # (bs, 256)

        candidate_embeddings = self.entity_embeddings(candidates_ids)  # (bs, num_candidates, 256)
        candidate_bias = self.entity_bias(candidates_ids).squeeze(-1)  # (bs, num_candidates

        logits = torch.bmm(
            span_embeds.unsqueeze(1),
            candidate_embeddings.transpose(1, 2),
        ).squeeze(1) + candidate_bias  # (bs, num_candidates)

        loss = None
        if labels is not None:
            loss = nn.functional.cross_entropy(logits, labels)

        return (loss, logits)

    def save_pretrained(self, save_directory: str) -> None:
        self.encoder.save_pretrained(save_directory)
        torch.save(self.projection.state_dict(), os.path.join(save_directory, "projection.pt"))
        torch.save(self.entity_embeddings.state_dict(), os.path.join(save_directory, 'embedding.pt'))
        torch.save(self.entity_bias.state_dict(), os.path.join(save_directory, 'bias.pt'))
        model_config = {
            "model_name_or_path": self.config["model_name_or_path"],
            "num_entities": self.config["num_entities"],
        }
        json.dump(model_config, open(os.path.join(save_directory, "model_config.json"), "w"), indent=2)

    @classmethod
    def from_pretrained(cls, load_directory: str) -> "SpanClassifier":
        model_config = json.load(open(os.path.join(load_directory, "model_config.json")))
        model_config["model_name_or_path"] = load_directory
        model = cls(**model_config)
        model.projection.load_state_dict(torch.load(os.path.join(load_directory, "projection.pt")))
        model.entity_embeddings.load_state_dict(torch.load(os.path.join(load_directory, "embedding.pt")))
        model.entity_bias.load_state_dict(torch.load(os.path.join(load_directory, 'bias.pt')))
        return model



import tempfile
from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers.trainer_utils import TrainOutput

import assets as test_data
from entity_linkings import get_rerankers, get_retrievers, load_dictionary
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import BaseSystemOutput

from .fusioned import FUSIONED

MODELS = ["google/flan-t5-base"]
dataset_path = str(files(test_data).joinpath("dataset_toy_wo_candidates.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)


@pytest.mark.reranker_fusioned
class TestFUSIONED:
    @pytest.mark.parametrize("use_checkpoint", [True, False])
    def test__init__(self, use_checkpoint: bool) -> None:
        model_cls = get_rerankers("fusioned")
        model = model_cls(retriever=retriever)
        assert isinstance(model, FUSIONED)
        assert hasattr(model, "config") and hasattr(model, "tokenizer") and hasattr(model, "dictionary")
        assert model.config.model_name_or_path == "google/flan-t5-base"
        assert model.config.max_context_length == 128
        assert model.config.max_candidate_length == 50
        assert model.config.context_window_chars == 500
        assert model.config.use_checkpoint == False if not use_checkpoint else True
        assert model.config.num_beams == 3
        assert model.config.max_new_tokens == 200
        assert model.config.min_length == 1
        assert model.config.document_token == "<extra_id_0>"
        assert model.config.passage_token == "<extra_id_1>"
        assert model.config.title_token == "<extra_id_2>"
        assert model.config.description_token == "<extra_id_3>"
        assert model.config.entity_token == "<extra_id_4>"
        assert model.config.mention_token == "<extra_id_5>"
        assert model.config.ent_start_token == "<extra_id_6>"
        assert model.config.ent_end_token == "<extra_id_7>"
        for d in model.dictionary:
            assert "encoding" in d

    @pytest.mark.parametrize("use_checkpoint", [True, False])
    def test_train(self, use_checkpoint: bool) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            model = FUSIONED(
                retriever=retriever,
                config=FUSIONED.Config(
                    model_name_or_path=MODELS[0],
                    use_checkpoint=use_checkpoint
                )
            )
            result = model.train(
                train_dataset=dataset,
                eval_dataset=dataset,
                num_candidates=3,
                training_args=TrainingArguments(
                    output_dir=tmpdir,
                    num_train_epochs=1,
                    per_device_train_batch_size=2,
                    per_device_eval_batch_size=2,
                    logging_strategy="no",
                    save_strategy="no",
                    eval_strategy="no",
                    remove_unused_columns=False,
                    eval_on_start=True
                )
            )
            assert isinstance(result, TrainOutput)
            assert hasattr(result, 'metrics')

    @pytest.mark.parametrize("use_checkpoint", [True, False])
    def test_evaluate(self, use_checkpoint: bool) -> None:
        model = FUSIONED(
            retriever=retriever,
            config=FUSIONED.Config(model_name_or_path=MODELS[0], use_checkpoint=use_checkpoint)
        )
        metrics = model.evaluate(dataset, num_candidates=3, batch_size=1)
        assert 'top1_accuracy' in metrics

    @pytest.mark.parametrize("use_checkpoint", [True, False])
    def test_predict(self, use_checkpoint: bool) -> None:
        sentence = "Apple is looking at buying U.K. startup for $1 billion"
        spans = [(0,5), (27,30), (44,52)]
        model = FUSIONED(
            retriever=retriever,
            config=FUSIONED.Config(model_name_or_path=MODELS[0], use_checkpoint=use_checkpoint)
        )
        predictions = model.predict(sentence, spans=spans, num_candidates=3)
        assert isinstance(predictions, list)
        for preds in predictions:
            assert isinstance(preds, BaseSystemOutput)
            assert preds.id is not None and preds.start is not None and preds.end is not None




from typing import Optional

import tiktoken
from openai import OpenAI
from tqdm.auto import tqdm

from .utils import DEFAULT_TIMEOUT, Cache


class OpenAI_API:
    """OpenAI_APIWrapper

    This wrapper is a inclusive OpenAI API tools.
    I strongly recommend to debug with estimate() at first, then use BatchAPI(ex submit_batch()) in experiment.
    """
    def __init__(
            self,
            model_name: str,
            token: str,
            organization_key: Optional[str] = None,
            max_generation_length: int = 4096,
            temperature: float = 0.2,
            top_p: float = 0.0,
            seed: int = 42,
            cache_dir: str = ".openai_cache",
        ) -> None:
        """__init__

        Args:
            model_name (str): OpenAI's model name
            api_file (str): Path of file of OpenAI api token. I assume that the api_token of OpenAI is formatting as json.
            total_cost_limit (float): Limit of total cost. This API terminates the process if the output cost is exceed of the limit. (Currency: USD)
            timeout (Optional[httpx.Timeout]): Timeout setting (Default: None)
        """
        self.client = OpenAI(organization=organization_key, api_key=token, timeout=DEFAULT_TIMEOUT)
        self.model_name = model_name
        self.temperature = temperature
        self.top_p = top_p
        self.seed = seed
        self.max_generation_tokens = max_generation_length
        self.cache = Cache(model_name=model_name, cache_dir=cache_dir)

    def _generate(self, prompt: str) -> str:
        try:
            response = self.client.chat.completions.create(
                model=self.model_name,
                messages=[ {"role": "user", "content": prompt} ],
                temperature=self.temperature,
                top_p=self.top_p,
                n=1,
                max_tokens=self.max_generation_tokens
            )
            generated_text = response.choices[0].message.content
            if not generated_text:
                raise ValueError("No response from OpenAI API.")
            return generated_text
        except Exception as e:
            return f"This process was terminated by the OpenAI API due to {str(e)}"

    def generate(self, prompts: str|list[str]) -> list[str]:
        generated_texts = []
        if isinstance(prompts, str):
            prompts = [prompts]
        pbar = tqdm(total=len(prompts), desc="Generating with OpenAI API")
        for prompt in prompts:
            pbar.update()
            if not self.cache.check_in_cache(prompt):
                _results = self._generate(prompt)
                results = self.cache.serialize(_results)
                self.cache.append_to_jsonl(prompt, results)
            else:
                results = self.cache(prompt)
            generated_texts.append(results)
        pbar.close()
        return generated_texts

    def estimate(self, prompts: str|list[str], estimated_output: str|list[str]) -> float:
        """Estimate cost of prompt

        Args:
            prompts (str|list[str]): prompt string or list of prompt strings

        Returns:
            float: estimated cost (Currency: USD)
        """
        tokenizer = tiktoken.encoding_for_model(self.model_name)
        prompt_tokens, generated_tokens = [], []
        if isinstance(prompts, str):
            assert isinstance(estimated_output, str)
            prompts = [prompts]
        if isinstance(estimated_output, str):
            estimated_outputs = [estimated_output]
        for prompt in prompts:
            prompt_tokens.extend(tokenizer.encode(prompt))
        for output in estimated_outputs:
            generated_tokens.extend(tokenizer.encode(output))
        cost = self._calculate_cost(self.model_name, len(prompt_tokens), len(generated_tokens) * len(prompts))
        return cost

    @staticmethod
    def _calculate_cost(model_name: str, num_prompt_tokens: int, num_generated_tokens: int = 0) -> float:
        cost_per_1k_tokens = {
            "gpt-3.5-turbo": {"prompt_tokens": 0.0005, "generated_tokens": 0.0015},
            "gpt-4.1": {"prompt_tokens": 0.002, "generated_tokens": 0.008},
            "gpt-4.1-mini": {"prompt_tokens": 0.0004, "generated_tokens": 0.0016},
            "gpt-4.1-nano": {"prompt_tokens": 0.0001, "generated_tokens": 0.0004},
            "gpt-4o": {"prompt_tokens": 0.0025, "generated_tokens": 0.01},
            "gpt-4o-mini": {"prompt_tokens": 0.00015, "generated_tokens": 0.0006}
        }
        if model_name not in cost_per_1k_tokens:
            raise ValueError(f"Cost for model {model_name} is not defined.")
        cost = 0.0
        cost += (num_prompt_tokens / 1000) * cost_per_1k_tokens[model_name]["prompt_tokens"]
        cost += (num_generated_tokens / 1000) * cost_per_1k_tokens[model_name]["generated_tokens"]
        return cost



import itertools
import json
import os
from argparse import ArgumentParser

from datasets import DownloadManager

_URLs = {
    "top": "https://huggingface.co/datasets/vera-pro/ShadowLink/resolve/main/Top.json",
    "shadow": "https://huggingface.co/datasets/vera-pro/ShadowLink/resolve/main/Shadow.json",
    "tail": "https://huggingface.co/datasets/vera-pro/ShadowLink/resolve/main/Tail.json",
}


def main(output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'shadowlink')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    uid = map(str, itertools.count(start=0, step=1))
    for subset_id, url in _URLs.items():
        data_path = DownloadManager().download_and_extract(url)
        with open(os.path.join(output_dir_data, f"{subset_id}.jsonl"), 'w', encoding='utf-8') as f:
            raw_examples = json.load(open(data_path, 'r', encoding='utf-8'))
            for i, r_e in enumerate(raw_examples):
                eid = next(uid)
                example = {
                    "subset": "shadowlink_"+subset_id,
                    "id": eid,
                    "text": r_e["example"],
                    "entities": [{
                        "start": r_e["span"][0],
                        "end": r_e["span"][0]+r_e["span"][1],
                        "label": [str(r_e["wiki_id"])]
                    }],
                }
                f.write(json.dumps(example, ensure_ascii=False) + '\n')


if __name__ == "__main__":
    parser = ArgumentParser(description="ShadowLink Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    args = parser.parse_args()

    main(args.output_dir)



from dataclasses import dataclass
from typing import Any, Optional, Union

import torch
from transformers import BatchEncoding
from transformers.data.data_collator import pad_without_fast_tokenizer_warning

from entity_linkings.data_utils import (
    CollatorBase,
    EntityDictionary,
)


@dataclass
class CollatorForCrossEncoder(CollatorBase):
    dictionary: Optional[EntityDictionary] = None
    train: bool = False

    def __call__(self, features: list[Union[dict[str, Any], BatchEncoding]]) -> BatchEncoding:
        if self.dictionary is None:
            raise ValueError("this modules needs to dictionary.")

        features = [f.copy() for f in features]
        candidates_encodings, batch_labels = [], []
        for f in features:
            context_tokens = f.pop("input_ids")
            candidate_ids = f.pop("candidates")
            candidates = [self.dictionary(cand) for cand in candidate_ids]

            gold_ids = f.pop("labels", None)
            if self.train:
                candidates = [self.dictionary(gold_ids[0])] + candidates[:-1]
                batch_labels.append(0)
            else:
                if gold_ids:
                    gold_label_ids = [1 if cand['id'] in gold_ids else 0 for cand in candidates]
                    if 1 not in gold_label_ids:
                        batch_labels.append(-1)
                    else:
                        batch_labels.append(gold_label_ids.index(1))

            for cand in candidates:
                candidate_tokens = cand['encoding']
                concat_tokens = context_tokens + candidate_tokens + [self.tokenizer.sep_token_id]
                cand_encodings = BatchEncoding({
                    "input_ids": concat_tokens,
                    "attention_mask": [1] * len(concat_tokens),
                })
                candidates_encodings.append(cand_encodings)

        batch = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            candidates_encodings,
            max_length=self.max_length,
            padding=True,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt"
        )

        for k, v in batch.items():
            cnum = len(v) // len(features)
            v = v.view(len(features), cnum, -1)
            batch[k] = v if self.return_tensors == "pt" else v.tolist()

        if batch_labels:
            batch["labels"] = torch.tensor(batch_labels) if self.return_tensors == "pt" else batch_labels
        return batch



import abc
from dataclasses import dataclass
from typing import Any, Optional

import numpy as np
from datasets import Dataset

from entity_linkings.data_utils import EntityDictionary
from entity_linkings.trainer import TrainingArguments
from entity_linkings.utils import BaseSystemOutput


class IndexerBase(abc.ABC):
    def __init__(self, dictionary: EntityDictionary) -> None:
        self.dictionary = dictionary
        self.entity_ids = self.dictionary.get_entity_ids()
        self.num_entities = len(self.dictionary)

    def _initialize(self) -> None:
        raise NotImplementedError

    def build_index(self, index_path: str) -> None:
        raise NotImplementedError

    def save_index(self, index_path: str, ensure_ascii: bool = False) -> None:
        raise NotImplementedError

    def load(self, index_path: str) -> None:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def search_knn(self, query: str|list[str], top_k: int, ignore_ids: Optional[list[str]|list[list[str]]] = None) -> tuple[np.ndarray, list[list[str]]]:
        raise NotImplementedError


class RetrieverBase(abc.ABC):
    '''
    Base class for entity retrieval models
    '''
    @dataclass
    class Config:
        model_name_or_path: Optional[str] = None
        def to_dict(self) -> dict[str, Any]:
            return self.__dict__

    def __init__(self, dictionary: EntityDictionary, config: Optional[Config] = None, index_path: Optional[str] = None) -> None:
        self.dictionary = dictionary
        self.config = config if config is not None else self.Config()

    def create_retriever(self, index_path: str) -> IndexerBase:
        raise NotImplementedError

    def train(self, train_dataset: Dataset, eval_dataset: Optional[Dataset] = None, num_hard_negatives: int = 0, training_args: Optional[TrainingArguments] = None) -> dict[str, float]:
        raise NotImplementedError

    def evaluate(self, dataset: Dataset, **args: int) -> dict[str, float]:
        raise NotImplementedError

    def predict(self, sentence: str, spans: Optional[list[tuple[int, int]]] = None, top_k: int = 5) -> list[list[BaseSystemOutput]]:
        raise NotImplementedError

    def retrieve_candidates(self, dataset: Dataset, top_k: int = 5, only_negative: bool = False, batch_size: int = 32, **args: int) -> list[list[str]]:
        raise NotImplementedError



from importlib.resources import files

from datasets import load_dataset
from torch.utils.data import DataLoader

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .collator import CollatorForRetrieval

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("dualencoder")
retriever = retriever_cls(dictionary=dictionary)
candidates_ids = retriever.retrieve_candidates(dataset, top_k=3, negative=True)

def test_CollatorForRetrieval() -> None:
    processed_dataset = retriever.preprocessor.dataset_preprocess(dataset, candidates_ids)
    collator = CollatorForRetrieval(retriever.tokenizer, dictionary=dictionary, num_hard_negatives=2)
    dataloader = DataLoader(processed_dataset, batch_size=2, collate_fn=collator)

    for batch in dataloader:
        assert "input_ids" in batch
        assert "attention_mask" in batch
        assert "token_type_ids" in batch
        assert "candidates_input_ids" in batch
        assert "candidates_attention_mask" in batch
        assert "candidates_token_type_ids" in batch
        assert "labels" in batch
        assert batch["input_ids"].size(0) == batch["attention_mask"].size(0) == batch["labels"].size(0) == 2
        assert batch["candidates_input_ids"].size(0) == batch["candidates_attention_mask"].size(0)
        assert batch["candidates_input_ids"].size(0) >= 2 and batch["candidates_input_ids"].size(0) <= 4
        assert batch['labels'].size() == (2, )



import abc
from typing import Optional

import spacy
from datasets import Dataset
from tqdm.auto import tqdm

from .candidate_reranker import RerankerBase
from .candidate_retriever import RetrieverBase
from .utils import BaseSystemOutput, calculate_inkb_f1


class ELPipeline(abc.ABC):
    def __init__(self, model: RetrieverBase | RerankerBase) -> None:
        self.model = model
        self.nlp = spacy.load("en_core_web_sm")

    def ner_predict(self, sentence: str) -> list[tuple[int, int]]:
        doc = self.nlp(sentence)
        spans = [(ent.start_char, ent.end_char) for ent in doc.ents]
        return spans

    def evaluate(self, dataset: Dataset, num_candidates: int = 30) -> dict[str, float]:
        pbar = tqdm(total=len(dataset), desc='Evaluate')
        all_predictions, all_golds = [], []
        for data in dataset:
            pbar.update()
            all_golds.append([ent for ent in data['entities'] if ent['label'] != ['-1']])
            # Mention detection
            text = data['text']
            pred_spans = self.ner_predict(text)
            if not pred_spans:
                all_predictions.append([])
                continue
            predictions = self.predict(text, spans=pred_spans, num_candidates=num_candidates)
            all_predictions.append([{'start': pred.start, 'end': pred.end, 'label': [pred.id]} for pred in predictions if pred.id != '-1'])
        pbar.close()
        metric = calculate_inkb_f1(all_predictions, all_golds)
        return metric

    def predict(self, sentence: str, spans: Optional[list[tuple[int, int]]] = None, num_candidates: int = 30) -> list[BaseSystemOutput]:
        if not spans:
            spans = self.ner_predict(sentence)
            if not spans:
                raise ValueError("No spans found using SpaCy. Please ensure the sentence contains named entities.")
        if isinstance(self.model, RetrieverBase):
            predictions = self.model.predict(sentence, spans, top_k=num_candidates)
            top1_predictions = [preds[0] for preds in predictions]
        elif isinstance(self.model, RerankerBase):
            top1_predictions = self.model.predict(sentence, spans, num_candidates=num_candidates)
        else:
            raise ValueError("The model should be either a RetrieverBase or RerankerBase instance.")
        return top1_predictions



import ast
import itertools
import json
import os
from argparse import ArgumentParser
from typing import Iterator, Optional

import pandas as pd
from datasets import DownloadManager
from tqdm.auto import tqdm

from appendix.data_convert.utils import WikiMapper

_URLs = {
    "train": [
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_0.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_1.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_3.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_4.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_5.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_6.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_7.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_8.csv.gz",
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_data/Tweeki_data_9.csv.gz",
    ],
    "test": [
        "https://github.com/ucinlp/tweeki/raw/refs/heads/main/data/Tweeki_gold/Tweeki_gold.jsonl"
    ]
}


def _convert(
        words: list[str],
        tags: list[str],
        word_list: list[str|list[str]],
        wiki_list: list[str],
        finder: Optional[WikiMapper] = None
    ) -> tuple[str, list[dict[str, str]]]:
    full_text = ' '.join(words)
    assert len(words) == len(tags) and len(word_list) == len(wiki_list)

    current_char_pos = 0
    token_offsets = []
    for word in words:
        start = current_char_pos
        end = start + len(word)
        token_offsets.append((start, end))
        current_char_pos = end + 1

    token_idx = 0
    entities = []
    for mention, label in zip(word_list, wiki_list):
        mention_token_len = len(mention) if mention else 1
        current_tag = tags[token_idx]

        wiki_id = "-1"
        if label != 0:
            wiki_title, qid = label.split('|')
            if finder is not None:
                res = finder.find_by_title(wiki_title)
                wiki_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
            else:
                wiki_id = qid

        if mention and current_tag.startswith('B-') or current_tag.startswith('U-'):
            start_offset = token_offsets[token_idx][0]
            end_offset = token_offsets[token_idx + mention_token_len - 1][1]
            assert full_text[start_offset:end_offset] == ' '.join(words[token_idx: token_idx + mention_token_len])

            entities.append({
                "start": start_offset,
                "end": end_offset,
                # "type": current_tag.split('-')[1],
                "label": [str(wiki_id)]
            })
        token_idx += mention_token_len

    return full_text, entities


def convert_csv_gz(input_path: str, finder: Optional[WikiMapper] = None) -> Iterator[tuple[str, list[dict[str, str]]]]:
    with open(input_path, 'rt', encoding='utf-8') as f:
        df = pd.read_csv(f)[['words', 'tags', 'word_list', 'wiki_list']]
        pbar = tqdm(total=len(df), desc="Processing csv.gz")
        for _, row in df.iterrows():
            pbar.update()
            words = ast.literal_eval(row['words'])
            tags = ast.literal_eval(row['tags'])
            words_list = ast.literal_eval(row['word_list'])
            wiki_list = ast.literal_eval(row['wiki_list'])

            full_text, entities = _convert(words, tags, words_list, wiki_list, finder)
            yield full_text, entities
        pbar.close()


def convert_jsonl(input_path: str, finder: Optional[WikiMapper] = None) -> Iterator[tuple[str, list[dict[str, str]]]]:
    with open(input_path, 'r', encoding='utf-8') as f:
        for line in f:
            data = json.loads(line)
            sentence = data['sentence']
            assert len(data['index']) == len(data['link'])

            entities = []
            for i in range(len(data['index'])):
                assert '|' in data['link'][i] and len(data['link'][i].split('|')) == 2
                title, qid = data['link'][i].split('|')
                title = title.replace(" ", "_")

                page_id = "-1"
                if finder is not None:
                    res = finder.find_by_title(title.strip())
                    page_id = res["page_id"] if res is not None else "-1"

                if page_id == "-1" and finder is not None:
                    res = finder.find_by_wikidata_id(qid.strip())
                    assert len(res) <= 1
                    page_id = res[0]["page_id"] if len(res) == 1 else "-1"

                assert isinstance(page_id, str)
                if page_id == "-1":
                    print(f"Warning: '{title}'({qid}) not found in wiki_map database.")

                entities.append({
                    "start": data['index'][i][0],
                    "end": data['index'][i][1],
                    "label": [str(page_id)]
                })

            yield sentence, entities


def convert(output_dir: str, finder: Optional[WikiMapper] = None) -> None:
    os.makedirs(output_dir, exist_ok=True)
    print(f"Output directory data: {output_dir}")

    uid = map(str, itertools.count(start=0, step=1))
    with open(os.path.join(output_dir, "train.jsonl"), 'w', encoding='utf-8') as f:
        for url in _URLs['train']:
            data_path = DownloadManager().download_and_extract(url)
            for text, entities in convert_csv_gz(data_path, finder):
                f.write(json.dumps({
                    "subset": "tweeki",
                    "id": next(uid),
                    "text": text,
                    "entities": entities
                }, ensure_ascii=False) + '\n')

    uid = map(str, itertools.count(start=0, step=1))
    with open(os.path.join(output_dir, "test.jsonl"), 'w', encoding='utf-8') as f:
        for url in _URLs['test']:
            data_path = DownloadManager().download_and_extract(url)
            for text, entities in convert_jsonl(data_path, finder):
                f.write(json.dumps({
                    "subset": "tweeki",
                    "id": next(uid),
                    "text": text,
                    "entities": entities
                }, ensure_ascii=False) + '\n')


def main(finder: WikiMapper, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'tweeki')
    os.makedirs(output_dir_data, exist_ok=True)

    convert(os.path.join(output_dir_data, 'wikidata'))
    convert(os.path.join(output_dir_data, 'wikipedia'), finder)


if __name__ == "__main__":
    parser = ArgumentParser(description="TweeKi Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, output_dir=args.output_dir)
    finder.close()



import os
from dataclasses import dataclass
from logging import getLogger
from typing import Optional

import torch
from datasets import Dataset
from torch.utils.data import DataLoader, SequentialSampler
from tqdm.auto import tqdm
from transformers import AutoModelForSeq2SeqLM, AutoTokenizer, set_seed

from entity_linkings.trainer import EntityLinkingTrainer, TrainingArguments
from entity_linkings.utils import BaseSystemOutput, calculate_top1_accuracy

from ..base import RerankerBase, RetrieverBase
from .collator import CollatorForFusioned
from .preprocessor import FusionedPreprocessor
from .reader import FusionDecoder

logger = getLogger(__name__)
logger.setLevel("INFO")


class FUSIONED(RerankerBase):
    """ Entity Disambiguation via Fusion Entity Decoding
    """

    @dataclass
    class Config(RerankerBase.Config):
        model_name_or_path: str = "google/flan-t5-base"
        max_context_length: int = 128
        max_candidate_length: int = 50
        context_window_chars: int = 500
        use_checkpoint: bool = False
        num_beams: int = 3
        max_new_tokens: int = 200
        min_length: int = 1
        document_token: str = "<extra_id_0>"
        passage_token: str = "<extra_id_1>"
        title_token: str = "<extra_id_2>"
        description_token: str = "<extra_id_3>"
        entity_token: str = "<extra_id_4>"
        mention_token: str = "<extra_id_5>"
        ent_start_token: str = "<extra_id_6>"
        ent_end_token: str = "<extra_id_7>"
        nil_token: str = "[NIL]"

    def __init__(self, retriever: RetrieverBase, config: Optional[Config] = None) -> None:
        super().__init__(retriever, config)
        self.tokenizer = AutoTokenizer.from_pretrained(self.config.model_name_or_path)
        if os.path.exists(self.config.model_name_or_path):
            self.model = FusionDecoder.from_pretrained(self.config.model_name_or_path)
        else:
            self.tokenizer.add_special_tokens({"additional_special_tokens": [self.config.nil_token]})
            t5 = AutoModelForSeq2SeqLM.from_pretrained(self.config.model_name_or_path)
            t5.resize_token_embeddings(len(self.tokenizer))
            self.model = FusionDecoder(t5.config)
            self.model.load_t5(t5.state_dict())
        self.model.set_checkpoint(self.config.use_checkpoint)

        self.preprocessor = FusionedPreprocessor(
            tokenizer = self.tokenizer,
            title_token=self.config.title_token,
            ent_start_token=self.config.ent_start_token,
            ent_end_token=self.config.ent_end_token,
            entity_token=self.config.entity_token,
            max_context_length=self.config.max_context_length,
            max_candidate_length=self.config.max_candidate_length,
            context_window_chars=self.config.context_window_chars,
        )
        self.dictionary = self.preprocessor.dictionary_preprocess(self.retriever.dictionary)

    def train(
        self,
        train_dataset: Dataset,
        eval_dataset: Optional[Dataset] = None,
        num_candidates: int = 30,
        training_args: Optional[TrainingArguments] = None
    ) -> dict[str, float]:
        if training_args is None:
            training_args = TrainingArguments()
        set_seed(training_args.seed)

        train_candidates = self.retriever.retrieve_candidates(
            train_dataset,
            only_negative=True,
            top_k=num_candidates,
            batch_size=training_args.per_device_eval_batch_size,
        )
        train_dataset = self.preprocessor.dataset_preprocess(train_dataset, train_candidates)

        if eval_dataset is not None:
            eval_candidates = self.retriever.retrieve_candidates(
                eval_dataset,
                only_negative=True,
                top_k=num_candidates,
                batch_size=training_args.per_device_eval_batch_size,
            )
            eval_dataset = self.preprocessor.dataset_preprocess(eval_dataset, eval_candidates)

        trainer = EntityLinkingTrainer(
            model=self.model,
            args=training_args,
            data_collator=CollatorForFusioned(
                self.tokenizer,
                dictionary=self.dictionary,
                train=True
            ),
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
        )
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint

        results = trainer.train(resume_from_checkpoint=checkpoint)
        trainer.log_metrics("train", results.metrics)
        if training_args.output_dir is not None:
            self.model.save_pretrained(training_args.output_dir)
            self.tokenizer.save_pretrained(training_args.output_dir)
            trainer.save_state()
            trainer.save_metrics("train", results.metrics)
        return results

    @torch.no_grad()
    def evaluate(self, dataset: Dataset, num_candidates: int = 30, batch_size: int = 32, **args: int) -> dict[str, float]:
        candidates = self.retriever.retrieve_candidates(
            dataset, only_negative=False, top_k=num_candidates, batch_size=batch_size,
        )

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.eval()
        self.model.to(device)
        processed_dataset = self.preprocessor.dataset_preprocess(dataset, candidates)

        dataloader = DataLoader(
            processed_dataset,
            batch_size=batch_size,
            sampler=SequentialSampler(processed_dataset),
            collate_fn=CollatorForFusioned(
                self.tokenizer,
                dictionary=self.dictionary,
                train=False,
            ),
        )

        num_corrects, num_golds = 0, 0
        pbar  = tqdm(total=len(dataloader), desc='Evaluate')
        for batch in dataloader:
            pbar.update()
            labels = batch.pop("labels") # (batch_size, n_candidates)
            batch = batch.to(device)
            generated_ids = self.model.generate(
                **batch,
                num_beams=self.config.num_beams,
                max_new_tokens=self.config.max_new_tokens,
                min_length=self.config.min_length,
            )
            generated_ids = generated_ids.to('cpu')
            for generated_id, label in zip(generated_ids, labels):
                gold = self.tokenizer.decode(label, skip_special_tokens=True)
                pred = self.tokenizer.decode(generated_id, skip_special_tokens=True)
                num_golds += 1
                num_corrects += 1 if pred == gold else 0
        pbar.close()
        metric = calculate_top1_accuracy(num_corrects, num_golds)
        return metric

    @torch.no_grad()
    def predict(self, sentence: str, spans: list[tuple[int, int]], num_candidates: int = 30) -> list[BaseSystemOutput]:
        predictions = self.retriever.predict(sentence, spans=spans, top_k=num_candidates)
        candidates = [[cand_id.id for cand_id in c] for c in predictions]

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.eval()
        self.model.to(device)

        all_result = []
        collator = CollatorForFusioned(self.tokenizer, dictionary=self.dictionary)
        for i, (b, e) in enumerate(spans):
            encoding = self.preprocessor.process_context(sentence, b, e, candidates[i])
            batch = collator([encoding])
            batch = batch.to(device)
            generated_ids = self.model.generate(
                **batch,
                num_beams=self.config.num_beams,
                max_new_tokens=self.config.max_new_tokens,
                min_length=self.config.min_length,
            )
            generated_ids = generated_ids.to('cpu')
            generated_text = self.tokenizer.decode(generated_ids[0], skip_special_tokens=True)

            candidate_titles = [self.dictionary(cand)['name'] for cand in candidates[i]]
            if generated_text not in candidate_titles:
                logger.warning(f"FUSIONED could not find a valid entity for the mention: {sentence[spans[i][0]:spans[i][1]]}")
                continue
            pred = candidate_titles.index(generated_text)
            top1_entity = self.dictionary(candidates[i][pred])
            all_result.append(BaseSystemOutput(query=sentence[b:e], start=b, end=e, id=top1_entity['id']))
        return all_result



import os
from dataclasses import dataclass
from typing import Optional

import torch
from datasets import Dataset
from torch.utils.data import DataLoader, SequentialSampler
from tqdm.auto import tqdm
from transformers import AutoTokenizer, set_seed

from entity_linkings.data_utils import Preprocessor
from entity_linkings.trainer import EntityLinkingTrainer, TrainingArguments
from entity_linkings.utils import BaseSystemOutput, calculate_top1_accuracy

from ..base import RerankerBase, RetrieverBase
from ..utils import compute_metrics
from .collator import CollatorForFEVRY
from .span_classifier import SpanClassifier


class FEVRY(RerankerBase):

    @dataclass
    class Config(RerankerBase.Config):
        '''FEVRY configuration
        '''
        model_name_or_path: str = "google-bert/bert-base-uncased"
        ent_start_token: str = "[START_ENT]"
        ent_end_token: str = "[END_ENT]"
        nil_token: str = "[NIL]"
        max_context_length: int = 128
        context_window_chars: int = 500

    def __init__(self, retriever: RetrieverBase, config: Optional[Config] = None) -> None:
        super().__init__(retriever, config)
        if os.path.exists(self.config.model_name_or_path):
            self.tokenizer = AutoTokenizer.from_pretrained(self.config.model_name_or_path)
            self.model = SpanClassifier.from_pretrained(self.config.model_name_or_path)
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(self.config.model_name_or_path)
            special_tokens = [self.config.ent_start_token, self.config.ent_end_token, self.config.nil_token]
            self.tokenizer.add_special_tokens({"additional_special_tokens": special_tokens})
            self.model = SpanClassifier(self.config.model_name_or_path, len(self.dictionary))
            self.model.resize_token_embeddings(len(self.tokenizer))
        self.preprocessor = Preprocessor(
            self.tokenizer,
            self.config.ent_start_token,
            self.config.ent_end_token,
            max_context_length=self.config.max_context_length,
            context_window_chars=self.config.context_window_chars
        )

    def train(
            self,
            train_dataset: Dataset,
            eval_dataset: Optional[Dataset] = None,
            num_candidates: int = 30,
            training_args: Optional[TrainingArguments] = None
        ) -> dict[str, float]:
        if training_args is None:
            training_args = TrainingArguments()
        set_seed(training_args.seed)

        train_candidates = self.retriever.retrieve_candidates(
            train_dataset,
            only_negative=True,
            top_k=num_candidates,
            batch_size=training_args.per_device_eval_batch_size,
        )
        train_dataset = self.preprocessor.dataset_preprocess(train_dataset, train_candidates)

        if eval_dataset is not None:
            eval_candidates = self.retriever.retrieve_candidates(
                eval_dataset,
                only_negative=True,
                top_k=num_candidates,
                batch_size=training_args.per_device_eval_batch_size,
            )
            eval_dataset = self.preprocessor.dataset_preprocess(eval_dataset, eval_candidates)

        trainer = EntityLinkingTrainer(
            model=self.model,
            args=training_args,
            data_collator=CollatorForFEVRY(self.tokenizer, dictionary=self.dictionary, train=True),
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=compute_metrics,
        )
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint

        results = trainer.train(resume_from_checkpoint=checkpoint)
        trainer.log_metrics("train", results.metrics)
        if training_args.output_dir is not None:
            self.model.save_pretrained(training_args.output_dir)
            self.tokenizer.save_pretrained(training_args.output_dir)
            trainer.save_state()
            trainer.save_metrics("train", results.metrics)
        return results

    @torch.no_grad()
    def evaluate(self, dataset: Dataset, num_candidates: int = 30, batch_size: int = 32, **args: int) -> dict[str, float]:
        candidates = self.retriever.retrieve_candidates(
            dataset, only_negative=False, top_k=num_candidates, batch_size=batch_size
        )

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.eval()
        self.model.to(device)

        processed_dataset = self.preprocessor.dataset_preprocess(dataset, candidates)
        dataloader = DataLoader(
            processed_dataset,
            batch_size=batch_size,
            sampler=SequentialSampler(processed_dataset),
            collate_fn=CollatorForFEVRY(self.tokenizer, dictionary=self.dictionary, train=False),
        )

        num_corrects, num_golds = 0, 0
        pbar  = tqdm(total=len(dataloader), desc='Evaluate')
        for batch in dataloader:
            pbar.update()
            batch = batch.to(device)
            labels = batch.pop("labels") # (batch_size, n_candidates)
            _, scores = self.model(**batch) # (batch_size, n_candidates)
            preds = scores.argmax(axis=1).ravel()  # (batch_size, )]
            labels = labels.ravel()
            mask = labels != -100
            preds = preds[mask]
            labels = labels[mask]
            num_corrects += (preds == labels).sum().item()
            num_golds += mask.sum().item()
        pbar.close()
        metric = calculate_top1_accuracy(num_corrects, num_golds)
        return metric

    @torch.no_grad()
    def predict(self, sentence: str, spans: list[tuple[int, int]], num_candidates: int = 30) -> list[BaseSystemOutput]:
        predictions = self.retriever.predict(sentence, spans=spans, top_k=num_candidates)
        candidates = [[cand_id.id for cand_id in c] for c in predictions]

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.eval()
        self.model.to(device)

        collator = CollatorForFEVRY(self.tokenizer, dictionary=self.dictionary, train=False)
        all_result = []
        for i, (b, e) in enumerate(spans):
            encodings = self.preprocessor.process_context(sentence, b, e, candidates[i])
            encodings = collator([encodings])
            encodings = encodings.to(device)
            _, scores = self.model(**encodings)
            preds = scores.argmax(axis=1)[0].item()
            top1_entity = self.dictionary(candidates[i][preds])
            all_result.append(BaseSystemOutput(query=sentence[b:e], start=b, end=e, id=top1_entity['id']))
        return all_result



import json
import logging
import os
from argparse import ArgumentParser, Namespace

import torch
from datasets import load_dataset

from entity_linkings import (
    ELPipeline,
    RerankerBase,
    RetrieverBase,
    get_rerankers,
    get_retrievers,
    load_dictionary,
)
from entity_linkings.utils import read_yaml

logger = logging.getLogger(__name__)

device = torch.device('cuda') if torch.cuda.is_available() else 'cpu'

def evaluate(args: Namespace) -> None:
    dictionary = load_dictionary(args.dictionary_id_or_path, cache_dir=args.cache_dir)
    dataset_id = args.dataset_id if args.dataset_id else "json"
    if dataset_id != "json":
        dataset_id = dataset_id if dataset_id.startswith("naist-nlp/") else f"naist-nlp/{dataset_id}"
        test_dataset = load_dataset(dataset_id, split='test', cache_dir=args.cache_dir)
    else:
        test_dataset = load_dataset("json", data_files={"test": args.test_file}, cache_dir=args.cache_dir)['test']
    if args.remove_nil:
        from entity_linkings.data_utils import filter_nil_entities
        test_dataset = filter_nil_entities(test_dataset, dictionary)

    if args.retriever_config is not None:
        retriever_config = read_yaml(args.retriever_config)[args.retriever_id.lower()]
    else:
        retriever_config = {}
    if args.retriever_model_name_or_path is not None:
        retriever_config["model_name_or_path"] = args.retriever_model_name_or_path
    retriever_cls = get_retrievers(args.retriever_id)
    retriever = retriever_cls(dictionary=dictionary, config=retriever_cls.Config(**retriever_config), index_path=args.retriever_index_dir)

    model: RerankerBase|RetrieverBase = retriever
    if args.reranker_id is not None:
        if args.reranker_config is not None:
            reranker_config = read_yaml(args.reranker_config).get(args.reranker_id, {})
        else:
            reranker_config = {}
        if args.reranker_model_name_or_path is not None:
            reranker_config["model_name_or_path"] = args.reranker_model_name_or_path
        reranker_cls = get_rerankers(args.reranker_id)
        reranker = reranker_cls(retriever, config=reranker_cls.Config(**reranker_config))
        model = reranker

    if args.wandb:
        import wandb
        wandb.init(
            project=os.environ.get("WANDB_PROJECT", "entity_linking_benchmark"),
            name=args.reranker_id, tags=["evaluation"]
        )
        wandb.log({
            "retriever_id": args.retriever_id,
            "reranker_model_name_or_path": reranker_config.get("model_name_or_path", None) if args.reranker_id is not None else None,
            "retriever_model_name_or_path": retriever_config.get("model_name_or_path", None),
            "remove_nil": args.remove_nil,
            "dataset_id": dataset_id,
            "test_file": args.test_file,
            "dictionary_id_or_path": args.dictionary_id_or_path,
        })

    pipeline = ELPipeline(model=model)
    metrics = pipeline.evaluate(test_dataset, num_candidates=args.num_candidates)

    logger.info(f"Evaluation results: {metrics}")
    if args.output_dir is not None:
        os.makedirs(args.output_dir, exist_ok=True)
        output_path = f"{args.output_dir}/eval_results.json"
        with open(output_path, "w") as f:
            json.dump(metrics, f, indent=4)
        logger.info(f"Saved evaluation results to {output_path}")
    if args.wandb:
        for key, value in metrics.items():
            wandb.log({key: value})


def cli_main() -> None:
    parser = ArgumentParser()
    parser.add_argument('--reranker_id', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--reranker_model_name_or_path', type=str, default=None, help='Name of the model to use.')
    parser.add_argument('--reranker_config', type=str, default=None, help='YAML-based reranker config file.')
    parser.add_argument('--retriever_id', type=str, required=True, help='Name of the retriever model to use.')
    parser.add_argument('--retriever_model_name_or_path', type=str, default=None, help='Name of the retriever model to use.')
    parser.add_argument('--retriever_index_dir', type=str, default=None, help='Path to the retriever index directory.')
    parser.add_argument('--retriever_config', type=str, default=None, help='YAML-based retriever config file.')
    parser.add_argument('--dictionary_id_or_path', '-d', type=str, default=None, help='Path to the entity dictionary file.')
    parser.add_argument('--dataset_id', '-D', type=str, default=None, help='Name of the dataset to use.')
    parser.add_argument('--test_file', type=str, default=None, help='Path to the dataset file.')
    parser.add_argument('--num_candidates', type=int, default=5, help='Number of candidate entities to consider during evaluation.')
    parser.add_argument('--test_batch_size', type=int, default=32, help='Batch size for evaluation.')
    parser.add_argument('--remove_nil', action='store_true', default=False, help='Whether to remove nil entities from the dataset.')
    parser.add_argument('--output_dir', type=str, default=None, help='Path to the output directory.')
    parser.add_argument("--cache_dir", type=str, default=None, help='Path to the cache directory.')
    parser.add_argument('--wandb', action='store_true', default=False, help='Whether to use wandb for logging.')
    args = parser.parse_args()
    evaluate(args)

if __name__ == "__main__":
    cli_main()



from importlib.resources import files

from datasets import load_dataset

import assets as test_data
from entity_linkings import load_dictionary

from .entity_dictionary import EntityDictionary

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))


class TestEntityDictionary:
    def test__init__(self) -> None:
        dictionary = load_dataset("json", data_files=dictionary_path, split="train")
        dictionary = EntityDictionary(dictionary)
        assert isinstance(dictionary, EntityDictionary)
        assert dictionary.cache_dir is None
        assert dictionary.nil_id == "-1"
        assert dictionary.nil_name == "<NIL>"
        assert dictionary.nil_description == "<NIL> is an entity that does not exist in this dictionary."
        assert dictionary.id_to_index == {"000011": 0, "000012": 1, "000013": 2, "000014": 3, "000015": 4, "-1": 5}
        assert isinstance(dictionary("000012"), dict) and dictionary("000012")["name"] == "Apple"
        assert isinstance(dictionary[0], dict) and dictionary[0]["name"] == "Google"
        assert len(dictionary) == 6

    def test_get_entity_ids(self) -> None:
        dictionary = load_dictionary(dictionary_path)
        ids = dictionary.get_entity_ids()
        assert isinstance(ids, list)
        assert ids == ["000011", "000012", "000013", "000014", "000015", "-1"]

    def test_get_entity_names(self) -> None:
        dictionary = load_dictionary(dictionary_path)
        names = dictionary.get_entity_names()
        assert isinstance(names, list)
        assert len(names) == len(dictionary)
        assert names[0] == "Google"
        assert names[1] == "Apple"
        assert names[2] == "Meta"
        assert names[3] == "Amazon"
        assert names[4] == "Microsoft"
        assert names[5] == "[NIL]"

    def test_get_entity_descriptions(self) -> None:
        dictionary = load_dictionary(dictionary_path)
        descriptions = dictionary.get_entity_descriptions()
        assert isinstance(descriptions, list)
        assert len(descriptions) == len(dictionary)
        assert descriptions[0] == "Google is a global company"
        assert descriptions[1] == "Apple is a global company"
        assert descriptions[2] == "Meta is a global company"
        assert descriptions[3] == "Amazon is a global company"
        assert descriptions[4] == "Microsoft is a global company"
        assert descriptions[5] == "[NIL] is an entity that does not exist in this dictionary."

    def test_convert_description(self) -> None:
        dictionary = load_dictionary(dictionary_path)
        description = dictionary.convert_description("TestEntity", "This is a test entity.")
        assert description == "This is a test entity."
        description = dictionary.convert_description("TestEntity", None)
        assert description == "TestEntity is an entity in this dictionary."

    def test_iter(self) -> None:
        dictionary = load_dictionary(dictionary_path)
        for i, entity in enumerate(dictionary):
            assert isinstance(entity, dict)




from importlib.resources import files

import pytest
from datasets import load_dataset
from torch.utils.data import DataLoader

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .collator import CollatorForExtend
from .extend import EXTEND

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)


@pytest.mark.parametrize("modify_global_attention", [1, 2])
def test_CollatorForExtend(modify_global_attention: int) -> None:
    model = EXTEND(retriever=retriever)
    candidates_ids = retriever.retrieve_candidates(dataset, top_k=3, only_negative=True)
    processed_dataset = model.preprocessor.dataset_preprocess(dataset, candidates_ids)
    collator = CollatorForExtend(model.tokenizer, modify_global_attention=modify_global_attention)
    dataloader = DataLoader(processed_dataset, batch_size=4, collate_fn=collator)

    for batch in dataloader:
        assert "input_ids" in batch
        assert "attention_mask" in batch
        assert batch["input_ids"].size(0) == batch["attention_mask"].size(0) == 4
        assert batch["input_ids"].size(1) == batch["attention_mask"].size(1)
        assert batch["global_attention_mask"].size() == batch["input_ids"].size()
        if "start_positions" in batch and "end_positions" in batch:
            print(batch["start_positions"].size(), batch["end_positions"].size())
            assert batch["start_positions"].size(0) == batch["end_positions"].size(0) == 4
        assert "labels" not in batch



import argparse
import gzip
import io
import re

import duckdb
import pandas as pd
import requests


def process_and_insert(con: duckdb.DuckDBPyConnection, url: str, table_name: str, pattern: re.Pattern) -> None:
    total_records = 0
    batch_size = 100000
    batch = []

    print(f"\nStreaming from URL: {url}")
    with (
        requests.get(url, stream=True, timeout=(10, 300)) as response,
        gzip.GzipFile(fileobj=response.raw) as gz,
        io.TextIOWrapper(gz, encoding='utf-8', errors='ignore') as f
    ):
        response.raise_for_status()
        for line in f:
            if line.startswith("INSERT INTO"):
                matches = pattern.findall(line)
                for m in matches:
                    processed_row = (int(m[0]), m[1].replace("\\'", "'"))
                    batch.append(processed_row)

                if len(batch) >= batch_size:
                    df_batch = pd.DataFrame(batch, columns=['page_id', 'title'])
                    con.execute(f"INSERT INTO {table_name} SELECT * FROM df_batch")
                    total_records += len(batch)
                    batch = []
                    print(f"Now inserting to {table_name}... Current total records: {total_records}")

    if batch:
        df_batch = pd.DataFrame(batch, columns=['page_id', 'title'])
        con.execute(f"INSERT INTO {table_name} SELECT * FROM df_batch")
        total_records += len(batch)
        print(f"Insertion complete to {table_name}. Total records: {total_records}")


def main(language: str, version: str, output_file: str) -> None:
    con = duckdb.connect(output_file)
    con.execute("CREATE OR REPLACE TABLE pages (page_id VARCHAR, title VARCHAR)")
    con.execute("CREATE OR REPLACE TABLE wikidata (page_id VARCHAR, wikidata_id VARCHAR)")
    con.execute("CREATE OR REPLACE TABLE redirects (page_id VARCHAR, rd_to VARCHAR)")

    base_url = f"https://dumps.wikimedia.org/{language}wiki/{version}/"
    page_file = f"{base_url}{language}wiki-{version}-page.sql.gz"
    props_file = f"{base_url}{language}wiki-{version}-page_props.sql.gz"
    redirect_file = f"{base_url}{language}wiki-{version}-redirect.sql.gz"

    try:
        print("Processing pages...")
        page_pattern = re.compile(r"\((\d+),0,'(.+?)',")
        process_and_insert(con, page_file, "pages", page_pattern)

        print("\nProcessing page properties...")
        props_pattern = re.compile(r"\((\d+),'wikibase_item','(Q\d+)',")
        process_and_insert(con, props_file, "wikidata", props_pattern)

        print("\nProcessing redirects...")
        redirect_pattern = re.compile(r"\((\d+),0,'(.*?)',")
        process_and_insert(con, redirect_file, "redirects", redirect_pattern)

        print("\nCreating final wiki_map table...")
        con.execute("""
            CREATE OR REPLACE TABLE wiki_map AS
                SELECT
                        p.page_id,
                        p.title,
                        self_props.wikidata_id AS wikidata_id,
                        CASE WHEN r.page_id IS NOT NULL THEN true ELSE false END AS is_redirect,
                        p_target.page_id as redirect_to
                FROM pages p
                LEFT JOIN wikidata self_props ON p.page_id = self_props.page_id
                LEFT JOIN redirects r ON p.page_id = r.page_id

                LEFT JOIN pages p_target ON r.rd_to = p_target.title
                LEFT JOIN wikidata target_props ON p_target.page_id = target_props.page_id
        """)

        # Dropping intermediate tables
        con.execute("DROP TABLE pages")
        con.execute("DROP TABLE wikidata")
        con.execute("DROP TABLE redirects")

        print("Creating indexes...")
        con.execute("CREATE INDEX title ON wiki_map (title)")
        con.execute("CREATE INDEX page_id ON wiki_map (page_id)")

        row_count = con.execute("SELECT count(*) FROM wiki_map").fetchone()[0]
        print("\n！""===============================")
        print(f"Created database: {output_file}")
        print(f"Total entities: {row_count:,}")

    except Exception as e:
        print(f"An error occurred: {e}")
    finally:
        con.close()



if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate wiki_map DuckDB database from Wikipedia dumps.")
    parser.add_argument('--language', '-l', type=str, default='en', help='Language code for Wikipedia dumps.')
    parser.add_argument('--version', '-v', type=str, default='latest', help='Version of Wikipedia dumps.')
    parser.add_argument('--output_file', '-o', type=str, default='wikipedia_to_wikidata.duckdb', help='Output DuckDB database file name.')

    args = parser.parse_args()
    main(args.language, args.version, args.output_file)



import tempfile

import pytest
import torch

from .fevry import SpanClassifier

MODELS = [
    "google-bert/bert-base-uncased",
    "FacebookAI/xlm-roberta-base",
    "microsoft/deberta-v3-base",
    "FacebookAI/roberta-base",
    "answerdotai/ModernBERT-base",
]

def mock_input(
        batch_size: int = 2,
        seq_length: int = 10,
        num_candidates: int = 3,
        num_entities: int = 10,
        vocab_size: int = 100
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    input_ids = torch.randint(0, vocab_size, (batch_size, seq_length))
    attention_mask = torch.ones((batch_size, seq_length), dtype=torch.long)
    start_positions = torch.randint(0, seq_length // 2, (batch_size,))
    end_positions = start_positions + torch.randint(1, seq_length // 2, (batch_size,))
    labels = torch.randint(0, num_candidates, (batch_size,))
    candidates = torch.randint(0, num_entities, (batch_size, num_candidates))
    return input_ids, attention_mask, start_positions, end_positions, labels, candidates


class TestSpanClassifier:
    @pytest.mark.parametrize("model_name", MODELS)
    def test_init_(self, model_name: str) -> None:
        encoder = SpanClassifier(model_name_or_path=model_name, num_entities=10)
        assert isinstance(encoder, SpanClassifier)
        assert encoder.projection.in_features == 1536
        assert encoder.projection.out_features == 256
        assert encoder.entity_embeddings.num_embeddings == 10
        assert encoder.entity_embeddings.embedding_dim == 256


    @pytest.mark.parametrize("model_name", MODELS)
    def test_encode(self, model_name: str) -> None:
        encoder = SpanClassifier(model_name_or_path=model_name, num_entities=10)
        input_ids, attention_mask, start_positions, end_positions, _, _ = mock_input()
        with torch.no_grad():
            span_embeds = encoder.encode(
                input_ids=input_ids,
                attention_mask=attention_mask,
                start_positions=start_positions,
                end_positions=end_positions,
            )
            assert span_embeds.size() == (input_ids.size(0), 1536)

    @pytest.mark.parametrize("model_name", MODELS)
    def test_forward(self, model_name: str) -> None:
        encoder = SpanClassifier(model_name_or_path=model_name, num_entities=10)
        input_ids, attention_mask, start_positions, end_positions, labels, candidates = mock_input()
        with torch.no_grad():
            _, logits = encoder(
                input_ids=input_ids,
                attention_mask=attention_mask,
                start_positions=start_positions,
                end_positions=end_positions,
                labels=labels,
                candidates_ids=candidates
            )
            assert logits.size() == (2, 3)

    def test_save_and_load_model(self) -> None:
        with tempfile.TemporaryDirectory() as tmpdir:
            encoder = SpanClassifier(model_name_or_path=MODELS[0], num_entities=20)
            encoder.save_pretrained(tmpdir)
            new_encoder = SpanClassifier.from_pretrained(tmpdir)
            assert new_encoder.projection.in_features == 1536
            assert new_encoder.projection.out_features == 256
            assert new_encoder.entity_embeddings.num_embeddings == 20
            assert new_encoder.entity_embeddings.embedding_dim == 256



from unittest.mock import MagicMock, patch

import pytest

from .gpt import OpenAI_API


class TestOpenAI_API:
    @pytest.fixture
    def api(self) -> OpenAI_API:
        return OpenAI_API(model_name="gpt-3.5-turbo", token="dummy-key")

    def test_generate(self, api: OpenAI_API) -> None:
        mock_response = MagicMock()
        mock_response.choices[0].message.content = "hello, world!"

        with patch.object(api.client.chat.completions, 'create', return_value=mock_response) as mock_create:
            result = api._generate("hello")
            assert result == "hello, world!"

            mock_create.assert_called_once_with(
                model="gpt-3.5-turbo",
                messages=[ {"role": "user", "content": "hello"} ],
                temperature=0.2,
                top_p=0.0,
                max_tokens=4096,
                n=1
            )

        with patch.object(api.client.chat.completions, 'create', side_effect=Exception("API Connection Error")):
            result = api._generate("hello")
            assert "This process was terminated by the OpenAI API" in result

    def test_estimate(self, api: OpenAI_API) -> None:
        prompts = ["Hello, how are you?", "What is the capital of France?"]
        cost = api.estimate(prompts, estimated_output="Fine.")
        assert isinstance(cost, float)



from importlib.resources import files

import pytest
from datasets import load_dataset

import assets as test_data
from entity_linkings import ELPipeline, get_rerankers, get_retrievers, load_dictionary
from entity_linkings.utils import BaseSystemOutput

from .candidate_reranker import RerankerBase
from .candidate_retriever import RetrieverBase

dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']

retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)
reranker_cls = get_rerankers("crossencoder")
reranker = reranker_cls(retriever=retriever)

class TestELPipeline:
    @pytest.mark.parametrize("model", [reranker, retriever])
    def test__init__(self, model: RerankerBase | RetrieverBase) -> None:
        pipeline = ELPipeline(model=model)
        assert isinstance(pipeline, ELPipeline)
        assert pipeline.model == model
        assert hasattr(pipeline, "nlp")

    def test_ner_predict(self) -> None:
        pipeline = ELPipeline(model=retriever)
        sentence = "Steve Jobs was founder of Apple."
        spans = pipeline.ner_predict(sentence)
        assert isinstance(spans, list)
        assert spans == [(0, 10), (26, 31)]

    @pytest.mark.parametrize("model", [reranker, retriever])
    def test_predict(self, model: RerankerBase | RetrieverBase) -> None:
        pipeline = ELPipeline(model=model)
        sentence = "Steve Jobs was founder of Apple."
        spans = [(0, 10), (26, 31)]
        predictions = pipeline.predict(sentence, spans=spans, num_candidates=5)
        assert isinstance(predictions, list)
        for pred in predictions:
            assert isinstance(pred, BaseSystemOutput)
            assert pred.id is not None and pred.start is not None and pred.end is not None

    @pytest.mark.parametrize("model", [reranker, retriever])
    def test_evaluate(self, model: RerankerBase | RetrieverBase) -> None:
        pipeline = ELPipeline(model=model)
        metric = pipeline.evaluate(dataset, num_candidates=5)
        assert isinstance(metric, dict)
        assert 'precision' in metric and 'recall' in metric and 'f1' in metric
        assert 0.0 <= metric['precision'] <= 1.0
        assert 0.0 <= metric['recall'] <= 1.0
        assert 0.0 <= metric['f1'] <= 1.0



from dataclasses import dataclass
from logging import getLogger
from typing import Optional

import torch
from datasets import Dataset

from entity_linkings.data_utils import cut_context_window
from entity_linkings.utils import BaseSystemOutput, calculate_top1_accuracy

from ..base import RerankerBase, RetrieverBase
from .gpt import OpenAI_API
from .utils import process_multi_choice_prompt

logger = getLogger(__name__)
logger.setLevel("INFO")


class CHATEL(RerankerBase):
    """ Entity Disambiguation via Fusion Entity Decoding
    """
    @dataclass
    class Config(RerankerBase.Config):
        model_name_or_path: str = "gpt-4o-mini"
        num_description_characters: int = 200
        max_generation_length: int = 100
        context_window_chars: int = 500
        temperature: float = 0.2
        top_p: float = 0.0
        seed: int = 0
        token: str = "<API_KEY>"
        organization_key: Optional[str] = None

    def __init__(self, retriever: RetrieverBase, config: Optional[Config] = None) -> None:
        super().__init__(retriever, config)
        self.model = OpenAI_API(
            model_name=self.config.model_name_or_path,
            temperature=self.config.temperature,
            top_p=self.config.top_p,
            seed=self.config.seed,
            max_generation_length=self.config.max_generation_length,
            token=self.config.token,
            organization_key=self.config.organization_key,
        )

    def first_step_prompt(self, sentence: str, start: int, end: int) -> str:
        mention = sentence[start:end]
        prompt_sentence, _, _ = cut_context_window(sentence, start, end, self.config.context_window_chars)
        return prompt_sentence + " \n What does " + mention + " in this sentence referring to?"

    def second_step_prompt(self, first_result: str, mention: str, candidates_ids: list[str]) -> str:
        multi_choice_prompt = ""
        candidates = [self.dictionary(c) for c in candidates_ids]
        for i, cand in enumerate(candidates):
            description = cand['name']
            if cand['description']:
                description += ' ' + cand['description'][:self.config.num_description_characters]
            multi_choice_prompt += f"({i+1}). {description}\n"
        multi_choice_prompt = first_result + '\n\n' + f'Which of the following does {mention} referes to?' + '\n\n' + multi_choice_prompt
        return multi_choice_prompt

    @torch.no_grad()
    def evaluate(self, dataset: Dataset, num_candidates: int = 30, batch_size: int = 32, **args: int) -> dict[str, float]:
        candidates = self.retriever.retrieve_candidates(
            dataset, only_negative=False, top_k=num_candidates, batch_size=batch_size,
        )

        first_prompts = []
        for text, entities in zip(dataset["text"], dataset["entities"]):
            for ent in entities:
                if not ent["label"]:
                    continue
                first_prompts.append(self.first_step_prompt(text, ent['start'], ent['end']))

        if isinstance(self.model, OpenAI_API):
            test_generation = self.model.generate(first_prompts[0])[0]
            estimated_cost = self.model.estimate(first_prompts, test_generation)
            logger.warning(f"Estimated cost for first step prompts: ${estimated_cost:.4f}")

        first_results = self.model.generate(first_prompts)

        second_prompts = []
        count = 0
        for text, entities in zip(dataset["text"], dataset["entities"]):
            for ent in entities:
                if not ent["label"]:
                    continue
                second_prompts.append(self.second_step_prompt(first_results[count], text[ent["start"]:ent["end"]], candidates[count]))
                count += 1

        if isinstance(self.model, OpenAI_API):
            test_generation = self.model.generate(second_prompts[0])[0]
            estimated_cost = self.model.estimate(second_prompts, test_generation)
            logger.warning(f"Estimated cost for second step prompts: ${estimated_cost:.4f}")

        second_results = self.model.generate(second_prompts)

        count = 0
        num_corrects, num_golds = 0, 0
        for entity in dataset["entities"]:
            for ent in entity:
                if not ent["label"]:
                    continue
                num_golds += 1
                result = second_results[count]
                choiced_id= process_multi_choice_prompt(result, candidates[count])
                pred = self.dictionary(candidates[count][choiced_id])['id'] if choiced_id != -1 else None
                if pred in ent["label"]:
                    num_corrects += 1
                count += 1
        return calculate_top1_accuracy(num_corrects, num_golds)

    @torch.no_grad()
    def predict(self, sentence: str, spans: list[tuple[int, int]], num_candidates: int = 30) -> list[BaseSystemOutput]:
        predictions = self.retriever.predict(sentence, spans, top_k=num_candidates)
        candidates = [[cand_id.id for cand_id in c] for c in predictions]

        first_prompts = []
        for b, e in spans:
            first_prompts.append(self.first_step_prompt(sentence, b, e))
        first_results = self.model.generate(first_prompts)

        second_prompts = []
        for i, (b, e) in enumerate(spans):
            second_prompts.append(self.second_step_prompt(first_results[i], sentence[b:e], candidates[i]))
        second_results = self.model.generate(second_prompts)

        all_result = []
        for i, result in enumerate(second_results):
            pred = process_multi_choice_prompt(result, candidates[i])
            if -1 == pred:
                logger.warning(f"CHATEL could not find a valid entity for the mention: {sentence[spans[i][0]:spans[i][1]]}")
                continue
            top1_entity = self.dictionary(candidates[i][pred])
            all_result.append(BaseSystemOutput(query=sentence[b:e], start=b, end=e, id=top1_entity['id']))
        return all_result



import bisect
import random
from typing import Any

from datasets import Dataset

from .entity_dictionary import EntityDictionary


def filter_nil_entities(dataset: Dataset, dictionary: EntityDictionary) -> Dataset:
    def filter_fn(example: dict[str, Any]) -> dict[str, Any]:
        entities = []
        for entity in example["entities"]:
            labels = [label for label in entity["label"] if dictionary(label) != dictionary.nil_id]
            if labels:
                entities.append({"start": entity["start"], "end": entity["end"], "label": labels})
        example.update({"entities": entities})
        return example
    return dataset.map(filter_fn)


def sample_range_excluding(n: int, k: int, excluding: list[int]) -> list[int]:
    skips = [j - i for i, j in enumerate(sorted(excluding))]
    s = random.sample(range(n - len(skips)), k)
    return [i + bisect.bisect_right(skips, i) for i in s]


def cut_context_window(text: str, start: int, end: int, context_window_chars: int) -> tuple[str, int, int]:
    mention_center = (start + end) // 2
    half_window = context_window_chars // 2

    window_start = max(0, mention_center - half_window)
    window_end = min(len(text), window_start + context_window_chars)

    if window_end - window_start < context_window_chars:
        window_start = max(0, window_end - context_window_chars)

    context = text[window_start: window_end]
    new_start = start - window_start
    new_end = end - window_start
    if new_start < 0 or new_end > len(context):
        raise ValueError("Adjusted mention span is out of context bounds.")
    return context, new_start, new_end


def truncate_around_mention(
    tokens_ids: list[int],
    max_token_length: int,
    ent_start_idx: int,
    ent_end_idx: int
) -> tuple[list[int], int, int]:
    if len(tokens_ids) <= max_token_length:
        return tokens_ids, ent_start_idx, ent_end_idx

    # Compute window around mention
    mention_center = (ent_start_idx + ent_end_idx) // 2
    half_window = max_token_length // 2

    left_context_length = mention_center
    right_context_length = len(tokens_ids) - mention_center

    if left_context_length >= half_window and right_context_length >= half_window:
        # Both sides have enough space → symmetric window
        left = mention_center - half_window
        right = mention_center + half_window
    elif left_context_length < half_window:
        # Not enough on the left → start from 0, extend right
        left = 0
        right = min(len(tokens_ids), max_token_length)
    else:  # right_space < half_window
        # Not enough on the right → end at len(text_tokens), extend left
        right = len(tokens_ids)
        left = max(0, right - max_token_length)
    truncated_text_tokens = tokens_ids[left:right]

    new_ent_start_idx = max(0, ent_start_idx - left)
    new_ent_end_idx = min(len(truncated_text_tokens), ent_end_idx - left)
    assert len(truncated_text_tokens) <= max_token_length

    return truncated_text_tokens, new_ent_start_idx, new_ent_end_idx



from .utils import process_multi_choice_prompt


def test_process_multi_choice_prompt() -> None:
    candidates = ["0001", "0002", "0003"]

    # Test case 1: Index selection
    result = process_multi_choice_prompt("(1).", [candidates[0]])
    assert result == 0  # Indexing starts from 0

    result = process_multi_choice_prompt("(2).", candidates)
    assert result == 1  # Indexing starts from 1



import logging
import random
from dataclasses import dataclass
from typing import Any, Optional, Union

import torch
from transformers import BatchEncoding
from transformers.data.data_collator import pad_without_fast_tokenizer_warning

from entity_linkings.data_utils import CollatorBase, EntityDictionary

logger = logging.getLogger(__name__)


@dataclass
class CollatorForRetrieval(CollatorBase):
    dictionary: Optional[EntityDictionary] = None
    num_hard_negatives: int = 0
    random_negative_sampling: bool = False

    def __call__(self, features: list[Union[dict[str, Any], BatchEncoding]]) -> BatchEncoding:
        if self.dictionary is None:
            raise ValueError("this modules needs to dictionary.")

        features = [f.copy() for f in features]
        inbatch_ids, hard_negative_ids = [], []
        for f in features:
            labels = f.pop("labels")
            inbatch_ids.append(labels[0])
            candidates = f.pop("candidates", None)
            if candidates is not None and self.num_hard_negatives > 0:
                if self.random_negative_sampling:
                    hard_negative_ids.extend(random.sample(candidates, self.num_hard_negatives))
                else:
                    hard_negative_ids.extend(candidates[:self.num_hard_negatives])

        batch = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors=self.return_tensors,
        )

        candidates_labels = list(set(inbatch_ids + list(set(hard_negative_ids))))
        candidates_encodings = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            [self.dictionary(cid)['encoding'] for cid in candidates_labels],
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors=self.return_tensors,
        )
        batch.update({f'candidates_{k}': v for k, v in candidates_encodings.items()})

        candidates_idx = {cid: idx for idx, cid in enumerate(candidates_labels)}
        inbatch_labels = [candidates_idx[cid] for cid in inbatch_ids]
        batch["labels"] = torch.tensor(inbatch_labels) if self.return_tensors == "pt" else inbatch_labels

        return batch



from typing import Any

import torch
import torch.nn as nn
from transformers import T5ForConditionalGeneration
from transformers.modeling_outputs import BaseModelOutput
from transformers.models.t5.configuration_t5 import T5Config
from transformers.models.t5.modeling_t5 import T5Stack


class FusionDecoder(T5ForConditionalGeneration):
    def __init__(self, config: T5Config) -> None:
        super().__init__(config)
        self.wrap_encoder()

    def forward(
            self,
            input_ids: torch.Tensor | None = None,
            attention_mask: torch.Tensor | None = None,
            **kwargs: Any
        ) -> tuple[Any, ...]:

        if input_ids is not None:
            if input_ids.dim() == 3:
                self.encoder.n_passages = input_ids.size(1)
            input_ids = input_ids.view(input_ids.size(0), -1)
        if attention_mask is not None:
            attention_mask = attention_mask.view(attention_mask.size(0), -1)

        return super(FusionDecoder, self).forward(
            input_ids=input_ids,
            attention_mask=attention_mask,
            **kwargs
        )

    def generate(
            self,
            input_ids: torch.Tensor,
            attention_mask: torch.Tensor,
            num_beams: int = 3,
            max_new_tokens: int = 200,
            min_length: int = 2
        ) -> torch.Tensor:
        self.encoder.n_passages = input_ids.size(1)
        return super().generate(
            input_ids=input_ids.view(input_ids.size(0), -1),
            attention_mask=attention_mask.view(attention_mask.size(0), -1),
            num_beams=num_beams,
            max_new_tokens=max_new_tokens,
            min_length=min_length
        )

    def wrap_encoder(self, use_checkpoint: bool =False) -> None:
        """
        Wrap T5 encoder to obtain a Fusion-in-Decoder model.
        """
        self.encoder = EncoderWrapper(self.encoder, use_checkpoint=use_checkpoint)

    def unwrap_encoder(self) -> None:
        """
        Unwrap Fusion-in-Decoder encoder, useful to load T5 weights.
        """
        self.encoder = self.encoder.encoder
        block = []
        for mod in self.encoder.block:
            block.append(mod.module)
        self.encoder.block = nn.ModuleList(block)

    def load_t5(self, state_dict: dict[str, torch.Tensor]) -> None:
        self.unwrap_encoder()
        self.load_state_dict(state_dict)
        self.wrap_encoder()

    def set_checkpoint(self, use_checkpoint: bool) -> None:
        assert isinstance(self.encoder, EncoderWrapper), "Encoder must be wrapped to set checkpointing."
        for mod in self.encoder.encoder.block:
            mod.use_checkpoint = use_checkpoint


class EncoderWrapper(torch.nn.Module):
    def __init__(self, encoder: T5Stack, use_checkpoint: bool = False) -> None:
        super().__init__()
        self.main_input_name = encoder.main_input_name
        self.encoder = encoder
        apply_checkpoint_wrapper(self.encoder, use_checkpoint)

    def forward(self, input_ids: torch.Tensor, attention_mask: torch.Tensor, **kwargs: Any) -> BaseModelOutput:
        bs, total_length = input_ids.size()
        passage_length = total_length // self.n_passages
        input_ids = input_ids.view(bs * self.n_passages, passage_length)
        attention_mask = attention_mask.view(bs * self.n_passages, passage_length)

        outputs = self.encoder(
            input_ids=input_ids,
            attention_mask=attention_mask,
            **kwargs
        )
        return BaseModelOutput(
            last_hidden_state=outputs[0].view(bs, self.n_passages * passage_length, -1),
            hidden_states=outputs[1:],
            attentions=None
        )


class CheckpointWrapper(nn.Module):
    def __init__(self, module: nn.Module, use_checkpoint: bool = False) -> None:
        super().__init__()
        self.module = module
        self.use_checkpoint = use_checkpoint

    def forward(
            self,
            hidden_states: dict[str, torch.Tensor],
            attention_mask: torch.Tensor | None = None,
            position_bias: torch.Tensor | None = None,
            encoder_hidden_states: torch.Tensor | None = None,
            encoder_attention_mask: torch.Tensor | None = None,
            encoder_decoder_position_bias: torch.Tensor | None = None,
            **kwargs: Any
        ) -> Any:
        if self.use_checkpoint and self.training:
            kwargs = {k: v for k, v in kwargs.items() if v is not None}
            def custom_forward(*inputs: Any) -> Any:
                output = self.module(*inputs, **kwargs)
                empty = torch.tensor(
                    [],
                    dtype=torch.float,
                    device=output[0].device,
                    requires_grad=True)
                output = tuple(x if x is not None else empty for x in output)
                return output

            output = torch.utils.checkpoint.checkpoint(
                custom_forward,
                hidden_states,
                attention_mask,
                position_bias,
                encoder_hidden_states,
                encoder_attention_mask,
                encoder_decoder_position_bias,
            )
            output = tuple(x if x.size() != 0 else None for x in output)
        else:
            output = self.module(
                hidden_states,
                attention_mask=attention_mask,
                position_bias=position_bias,
                encoder_hidden_states=encoder_hidden_states,
                encoder_attention_mask=encoder_attention_mask,
                encoder_decoder_position_bias=encoder_decoder_position_bias,
                **kwargs,
            )
        return output


def apply_checkpoint_wrapper(t5stack: T5Stack, use_checkpoint: bool) -> None:
    """
    Wrap each block of the encoder to enable checkpointing.
    """
    block = []
    for mod in t5stack.block:
        wrapped_mod = CheckpointWrapper(mod, use_checkpoint)
        block.append(wrapped_mod)
    t5stack.block = nn.ModuleList(block)



import html
import json
import os
import re
import unicodedata
from argparse import ArgumentParser
from typing import Any
from xml.etree import ElementTree

from tqdm.auto import tqdm

from appendix.data_convert.utils import WikiMapper


def create_unified_normalization_map(raw_text: str) -> tuple[str, list[int]]:
    intermediate_chars = []
    pattern = re.compile(r"&[a-zA-Z0-9#]+;")
    last_idx = 0

    for match in pattern.finditer(raw_text):
        for i in range(last_idx, match.start()):
            intermediate_chars.append((raw_text[i], i))

        decoded = html.unescape(match.group())
        for char in decoded:
            intermediate_chars.append((char, match.start()))
        last_idx = match.end()

    for i in range(last_idx, len(raw_text)):
        intermediate_chars.append((raw_text[i], i))

    final_text = ""
    old_to_new = [0] * (len(raw_text) + 1)

    current_final_idx = 0
    for char, original_idx in intermediate_chars:
        normalized_char = unicodedata.normalize('NFKC', char)
        final_text += normalized_char

        target_orig_idx = original_idx
        while len(old_to_new) > target_orig_idx and target_orig_idx <= original_idx:
            target_orig_idx += 1

        for i in range(original_idx, len(raw_text) + 1):
            old_to_new[i] = current_final_idx

        current_final_idx += len(normalized_char)

    old_to_new[len(raw_text)] = current_final_idx

    return final_text, old_to_new


def _convert(input_dir: str, finder: WikiMapper, subset: str) -> list[dict[str, Any]]:
    data_dir = os.path.join(input_dir, "wned-datasets", subset)
    files = [fn for fn in os.listdir(os.path.join(data_dir, "RawText"))]
    raw_texts = {}
    index_maps = {}
    pbar = tqdm(total=len(files), desc="Preprocessing raw texts")
    for fn in files:
        pbar.update(1)
        context = open(os.path.join(data_dir, "RawText", fn), 'r', encoding='utf-8').read()
        context, index_map = create_unified_normalization_map(context)
        raw_texts[fn] = context
        index_maps[fn] = index_map
    pbar.close()

    xml_file = os.path.join(data_dir, "clueweb.xml") if subset == 'clueweb12' else os.path.join(data_dir, "wikipedia.xml")

    examples = []
    tree = ElementTree.parse(xml_file)
    root = tree.getroot()

    documents = root.findall("document")
    pbar = tqdm(total=len(documents), desc="Processing documents")
    for elem in documents:
        pbar.update()
        entities = []
        docname = elem.attrib['docName']

        plus_char_num = 0
        for child in elem.findall("annotation"):
            wikiname = child[1].text.replace(' ', '_')
            res = finder.find_by_title(wikiname)
            page_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
            pageid = str(page_id)

            start = int(child[2].text)
            end = start + int(child[3].text)
            new_start, new_end = index_maps[docname][start], index_maps[docname][end]
            if child[0].text != raw_texts[docname][new_start:new_end]:
                if docname == 'clueweb12-0503wb-01-33875':
                    if child[0].text == 'Peter Paige' and raw_texts[docname][new_start + plus_char_num: new_end + plus_char_num] == '(Peter Paig':
                        plus_char_num += 1
                    if child[0].text == 'Pittsburgh':
                        if raw_texts[docname][new_start + plus_char_num: new_end + plus_char_num] == ' Pittsburg':
                            plus_char_num += 1
                        if raw_texts[docname][new_start + plus_char_num: new_end + plus_char_num] == 'ittsburrgh.':
                            plus_char_num -= 1
                    if child[0].text == 'QUEER AS FOLK' and raw_texts[docname][new_start + plus_char_num: new_end + plus_char_num] == 's\nQUEER AS FO':
                        plus_char_num += 2
                    new_start += plus_char_num
                    new_end += plus_char_num
            assert child[0].text == raw_texts[docname][new_start:new_end], f"{child[0].text} != {raw_texts[docname][new_start:new_end]}"
            entities.append({"start": new_start, "end": new_end, "label": [pageid]})

        examples.append({
            "subset": subset,
            "id": docname,
            "text": raw_texts[docname],
            "entities": entities,
        })
    pbar.close()
    return examples


def main(input_dir: str, output_dir: str, finder: WikiMapper) -> None:
    output_dir_data = os.path.join(output_dir, 'wned')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    examples = _convert(input_dir, finder, subset="wikipedia")
    with open(os.path.join(output_dir_data, "wikipedia.jsonl"), 'w', encoding='utf-8') as f:
        for item in examples:
            f.write(json.dumps(item, ensure_ascii=False) + '\n')

    examples = _convert(input_dir, finder, subset="clueweb12")
    with open(os.path.join(output_dir_data, "clueweb.jsonl"), 'w', encoding='utf-8') as f:
        for item in examples:
            f.write(json.dumps(item, ensure_ascii=False) + '\n')


if __name__ == "__main__":
    parser = ArgumentParser(description="WNED Entity Disambiguation Script")
    parser.add_argument("--input_dir", "-i", type=str, required=True, help="Path to the input text file")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(args.input_dir, args.output_dir, finder)
    finder.close()



from importlib.resources import files

import pytest
from datasets import load_dataset

import assets as test_data
from entity_linkings import load_dictionary
from entity_linkings.utils import BaseSystemOutput

from .prior import PRIOR

dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
mention_counter_path = str(files(test_data).joinpath("mention_counter_toy.json"))


@pytest.fixture(scope='module')
def prior_model() -> PRIOR:
    model = PRIOR(
        dictionary=dictionary,
        config=PRIOR.Config(model_name_or_path=mention_counter_path)
    )
    return model


@pytest.mark.retriever_prior
class TestPrior:
    def test_init(self, prior_model: PRIOR) -> None:
        assert isinstance(prior_model, PRIOR)
        assert prior_model.retriever is not None
        assert prior_model.dictionary is not None

    def test_evaluate(self, prior_model: PRIOR) -> None:
        metrics = prior_model.evaluate(dataset)
        assert 'recall@1' in metrics
        assert 'recall@10' in metrics
        assert 'recall@50' in metrics
        assert 'recall@100' in metrics
        assert 'mrr@100' in metrics

    def test_predict(self, prior_model: PRIOR) -> None:
        sentence = "Steve Jobs was found Apple."
        spans = [(21, 26)]
        top_k = 3
        predictions = prior_model.predict(sentence, spans, top_k)
        assert isinstance(predictions, list)
        assert len(predictions) == len(spans)
        for preds in predictions:
            assert isinstance(preds, list)
            assert len(preds) == min(top_k, len(dictionary))
            for pred in preds:
                assert isinstance(pred, BaseSystemOutput)
                assert pred.id is not None and pred.start is not None and pred.end is not None


    def test_retrieve_candidates(self, prior_model: PRIOR) -> None:
        top_k = 5
        candidates = prior_model.retrieve_candidates(dataset, top_k=top_k, negative=False, batch_size=3)
        assert isinstance(candidates, list)
        total_entities = sum(len(example["entities"]) for example in dataset)
        assert len(candidates) == total_entities
        for inds in candidates:
            assert isinstance(inds, list)
            assert len(inds) == top_k



import json
import os
from argparse import ArgumentParser

from datasets import DownloadManager

from appendix.data_convert.utils import WikiMapper, read_text_file

_URL = "https://community.nzdl.org/wikification/data/wikifiedStories.zip"



def main(finder: WikiMapper, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'aquaint')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    data_dir = os.path.join(DownloadManager().download_and_extract(_URL))
    file_names = [fn for fn in os.listdir(data_dir) if fn.endswith('.htm')]

    with open(os.path.join(output_dir_data, "test.jsonl"), 'w', encoding='utf-8') as f:
        for i, file_name in enumerate(file_names):
            raw_text = read_text_file(os.path.join(data_dir, file_name))
            texts = raw_text.split('<p> ')
            texts[0] = texts[0].split('<h1 id="header">')[1][:-7]
            for i in range(1, len(texts) - 1):
                texts[i] = texts[i][:-7]
            texts[-1] = texts[-1][:-23]

            char_len = 0
            procesed_texts = []
            entities = []
            for text in texts:
                if not text.strip():
                    continue

                current_entity = text.find("[[")
                while current_entity != -1:
                    wikiname = ""
                    surface = ""
                    j = current_entity + 2

                    while text[j] not in ["]", "|"]:
                        wikiname += text[j]
                        j += 1

                    if text[j] == "]":
                        surface = wikiname
                    else:
                        j += 1
                        while text[j] not in ["]", "|"]:
                            surface += text[j]
                            j += 1

                        if text[j] =="|":
                            agreement_score = float(text[j+1: j+4])
                            j += 4
                            if agreement_score < 0.5:
                                text = text[:current_entity] + surface + text[j+2:]
                                current_entity = text.find("[[")
                                continue

                    title = wikiname[0].upper() + wikiname.replace(" ", "_")[1:]
                    res = finder.find_by_title(title)
                    page_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
                    entities.append({
                        "start": current_entity + char_len,
                        "end": current_entity + char_len + len(surface),
                        "label": [str(page_id)],
                    })

                    text = text[:current_entity] + surface + text[j+2:]
                    assert surface == text[current_entity: current_entity + len(surface)]
                    current_entity = text.find("[[")
                procesed_texts.append(text)
                char_len += len(text) + 1  # +1 for '\n'

            example = {
                "subset": "aquaint",
                "id": file_name[:-4],
                "text": '\n'.join(procesed_texts),
                "entities": entities,
            }
            f.write(f"{json.dumps(example, ensure_ascii=False)}\n")


if __name__ == "__main__":
    parser = ArgumentParser(description="AQUAINT Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, output_dir=args.output_dir)
    finder.close()



import json
import os
from argparse import ArgumentParser
from itertools import chain

from datasets import DownloadManager
from tqdm.auto import tqdm

from appendix.data_convert.utils import WikiMapper

_URLs = {
    "train": "https://raw.githubusercontent.com/alteca/Linked-DocRED/refs/heads/main/Linked-Re-DocRED/train_revised.json",
    "validation": "https://raw.githubusercontent.com/alteca/Linked-DocRED/refs/heads/main/Linked-Re-DocRED/dev_revised.json",
    "test": "https://raw.githubusercontent.com/alteca/Linked-DocRED/refs/heads/main/Linked-Re-DocRED/test_revised.json"
}



def _convert(output_dir: str, finder: WikiMapper, source: str) -> None:
    output_dir_data = os.path.join(output_dir, f'docred/{source}')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    for split_name, url in _URLs.items():
        ds = DownloadManager().download_and_extract(url)
        raw_examples = json.load(open(ds, 'r', encoding='utf-8'))

        output_path = os.path.join(output_dir_data, f"{split_name}.jsonl")
        with open(output_path, 'w', encoding='utf-8') as f:
            pbar = tqdm(total=len(raw_examples), desc=f"Processing {split_name}")
            for r_e in raw_examples:
                pbar.update(1)
                sentences = r_e['sents']
                global_sentences = list(chain.from_iterable(sentences))
                entities = []
                for entity in r_e['entities']:
                    entity_info = entity['entity_linking']
                    if entity_info['confidence'] != 'A':
                        continue

                    resource = entity_info[f'{source}_resource']
                    if resource in ['#ignored#', f"#DocRED-{entity['id']}"]:
                        page_id = "-1"
                    else:
                        if source == "wikipedia":
                            res = finder.find_by_title(resource)
                            page_id = str(res["page_id"]) if res is not None and res["page_id"] is not None else "-1"
                        elif source == "wikidata":
                            page_id = resource

                    for mention in entity["mentions"]:
                        sent_id = mention['sent_id']
                        start, end = mention['pos'][0], mention['pos'][1]
                        char_start = len(" ".join(sentences[sent_id][:start])) + (1 if start > 0 else 0)
                        char_end = len(" ".join(sentences[sent_id][:end]))
                        char_context = len(" ".join(list(chain.from_iterable(sentences[: sent_id]))))
                        global_char_start = char_start + char_context + (1 if char_context > 0 else 0)
                        global_char_end = char_end + char_context + (1 if char_context > 0 else 0)
                        entities.append({
                            "start": global_char_start,
                            "end": global_char_end,
                            "label": [page_id],
                        })

                example = {
                    "subset": "linked_re-docred",
                    "id": f"{r_e['title']}",
                    "text": " ".join(global_sentences),
                    "entities": entities,
                }
                f.write(json.dumps(example, ensure_ascii=False) + "\n")
            pbar.close()


def main(finder: WikiMapper, output_dir: str) -> None:
    _convert(output_dir, finder, source="wikipedia")
    _convert(output_dir, finder, source="wikidata")


if __name__ == "__main__":
    parser = ArgumentParser(description="Linked-DocRed Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, output_dir=args.output_dir)
    finder.close()



import logging
import random
from dataclasses import dataclass
from typing import Any, Optional, Union

from transformers import BatchEncoding
from transformers.data.data_collator import pad_without_fast_tokenizer_warning

from entity_linkings.data_utils import CollatorBase, EntityDictionary

logger = logging.getLogger(__name__)



@dataclass
class CollatorForFusioned(CollatorBase):
    dictionary: Optional[EntityDictionary] = None
    train: bool = False

    def __call__(self, features: list[Union[dict[str, Any], BatchEncoding]]) -> BatchEncoding:
        if self.dictionary is None:
            raise ValueError("this modules needs to dictionary.")

        features = [f.copy() for f in features]
        new_features, batch_labels = [], []
        for f in features:
            context_tokens = f.pop("input_ids")
            candidates_ids = f.pop("candidates")
            candidates = [self.dictionary(cand) for cand in candidates_ids]

            golds = f.pop("labels", None)
            if golds is not None:
                gold_entity = self.dictionary(golds[0])
                batch_labels.append(gold_entity['name'])
                if self.train:
                    candidates = [gold_entity] + candidates[:-1]
                    random.shuffle(candidates)

            for cand in candidates:
                encoding = self.tokenizer.prepare_for_model(
                    context_tokens + cand['encoding'],
                    truncation=True,
                    add_special_tokens=True,
                )
                new_features.append(encoding)

        batch = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            new_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )

        for k, v in batch.items():
            cnum = len(v) // len(features)
            v = v.view(len(features), cnum, -1)
            batch[k] = v if self.return_tensors == "pt" else v.tolist()

        if batch_labels:
            batch['labels'] = self.tokenizer(
                batch_labels,
                padding=True,
                truncation=True,
                return_tensors=self.return_tensors,
            ).input_ids
        return batch



import torch

from .utils import (
    calculate_joint_probabilities,
    select_indices,
)


def test_calculate_joint_probabilities() -> None:
    classification_probabilities = torch.tensor([
        [0.1, 0.3, 0.6],
        [0.3, 0.2, 0.5],
    ])
    possible_indices = [(0, 2), (1, 1), (2, 0)]

    max_prod_probs = calculate_joint_probabilities(
        "max-prod",
        possible_indices,
        classification_probabilities
    )
    assert max_prod_probs == [
        torch.tensor(0.1) * torch.tensor(0.5),
        torch.tensor(0.3) * torch.tensor(0.2),
        torch.tensor(0.6) * torch.tensor(0.3),
    ]

    max_start_probs = calculate_joint_probabilities(
        "max-start",
        possible_indices,
        classification_probabilities
    )
    assert max_start_probs == [torch.tensor(0.1), torch.tensor(0.3), torch.tensor(0.6)]


    max_end_probs = calculate_joint_probabilities(
        "max-end",
        possible_indices,
        classification_probabilities
    )
    assert max_end_probs == [torch.tensor(0.5), torch.tensor(0.2), torch.tensor(0.3)]

    max_end_probs = calculate_joint_probabilities(
        "max",
        possible_indices,
        classification_probabilities
    )
    assert max_end_probs == [torch.tensor(0.5), torch.tensor(0.3), torch.tensor(0.6)]


def test_select_indices() -> None:
    classification_probabilities = torch.tensor([
        [0.1, 0.3, 0.6],
        [0.3, 0.2, 0.5],
    ])
    possible_indices = [(0, 2), (1, 1), (2, 0)]
    start_idx, end_idx = select_indices(
        "max-prod",
        possible_indices,
        classification_probabilities
    )
    assert (start_idx, end_idx) == (2, 0)

    start_idx, end_idx = select_indices(
        "max-start",
        possible_indices,
        classification_probabilities
    )
    assert (start_idx, end_idx) == (2, 0)

    start_idx, end_idx = select_indices(
        "max-end",
        possible_indices,
        classification_probabilities
    )
    assert (start_idx, end_idx) == (0, 2)

    start_idx, end_idx = select_indices(
        "max",
        possible_indices,
        classification_probabilities
    )
    assert (start_idx, end_idx) == (2, 0)



from importlib.resources import files

import pytest
from datasets import load_dataset
from torch.utils.data import DataLoader

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .collator import CollatorForFusioned
from .fusioned import FUSIONED

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']
retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)


@pytest.mark.parametrize("train", [True, False])
def test_CollatorForFusioned(train: bool) -> None:
    model = FUSIONED(retriever=retriever)
    candidates_ids = retriever.retrieve_candidates(dataset, top_k=3, only_negative=True if train else False)
    processed_dataset = model.preprocessor.dataset_preprocess(dataset, candidates_ids)
    collator = CollatorForFusioned(model.tokenizer, dictionary=dictionary, train=train)
    dataloader = DataLoader(processed_dataset, batch_size=2, collate_fn=collator)

    for batch in dataloader:
        assert "input_ids" in batch
        assert "attention_mask" in batch
        assert "labels" in batch
        assert batch["input_ids"].size(0) == batch["attention_mask"].size(0) == 2
        assert batch["input_ids"].size(1) == batch["attention_mask"].size(1) == 3
        assert batch["labels"].size(0) == 2



import logging
import math
import random
from dataclasses import dataclass
from typing import Literal, Optional

from datasets import Dataset
from tqdm.auto import tqdm

from entity_linkings.data_utils import EntityDictionary
from entity_linkings.utils import BaseSystemOutput, calculate_recall_mrr

from ..base import RetrieverBase
from .indexer import BM25Indexer

logger = logging.getLogger(__name__)


class BM25(RetrieverBase):
    '''
    BM25 model for entity disambiguation
    '''
    @dataclass
    class Config(RetrieverBase.Config):
        language: str = "en"
        n_threads: int = -1
        subword_tokenizer: Optional[str] = None
        query_type_for_candidate: Literal['mention', 'description'] = 'mention'

    def __init__(self, dictionary: EntityDictionary, config: Optional[Config] = None, index_path: Optional[str] = None) -> None:
        super().__init__(dictionary, config)
        if index_path is not None:
            self.retriever = self.create_retriever(index_path=index_path)

    def create_retriever(self, index_path: str) -> BM25Indexer:
        retriever =  BM25Indexer(
            dictionary=self.dictionary,
            language=self.config.language,
            n_threads=self.config.n_threads,
            subword_tokenizer=self.config.subword_tokenizer,
            query_type_for_candidate=self.config.query_type_for_candidate
        )
        retriever.build_index(index_path=index_path)
        return retriever

    def evaluate(self, dataset: Dataset, batch_size: int = 32, **args: int) -> dict[str, float]:
        queries, labels = [], []
        for text, entities in zip(dataset["text"], dataset["entities"]):
            for ent in entities:
                ent_labels = ent['label']
                if not ent_labels:
                    continue
                if self.config.query_type_for_candidate == 'mention':
                    queries.append(text[ent["start"]: ent["end"]])
                else:
                    ent_label = random.choice(ent_labels)
                    queries.append(self.dictionary(ent_label)["description"])
                labels.append(ent_labels)

        predictions = []
        pbar = tqdm(total=(math.ceil(len(queries)/batch_size)), desc='Evaluate')
        for i in range(0, len(queries), batch_size):
            pbar.update()
            _, batch_indices = self.retriever.search_knn(queries[i:i + batch_size], top_k=100)
            batch_labels = labels[i:i + batch_size]
            for j, indices in enumerate(batch_indices):
                preds = [{"id": self.dictionary(inds)["id"]} for inds in indices]
                predictions.append({"gold": batch_labels[j], "predict": preds})
        pbar.close()
        metric = calculate_recall_mrr(predictions)
        return metric

    def predict(self, sentence: str, spans: Optional[list[tuple[int, int]]] = None, top_k: int = 5) -> list[list[BaseSystemOutput]]:
        if not spans:
            raise ValueError("Spans must be provided for BM25 prediction.")

        queries = []
        for b, e in spans:
            queries.append(sentence[b:e])
        similarities, indices = self.retriever.search_knn(queries, top_k=top_k)
        all_result = []
        for i, (b, e) in enumerate(spans):
            result = []
            query = queries[i]
            for _, ind in enumerate(indices[i]):
                entry = self.dictionary(ind)
                result.append(BaseSystemOutput(query=query, start=b, end=e, id=entry['id']))
            all_result.append(result)
        return all_result

    def retrieve_candidates(self, dataset: Dataset, top_k: int = 5, only_negative: bool = False, batch_size: int = 32, **args: int) -> list[list[str]]:
        queries, labels = [], []
        for example in dataset:
            text = example['text']
            for ent in example["entities"]:
                ent_labels = ent['label']
                if not ent_labels:
                    continue
                if self.config.query_type_for_candidate == 'mention':
                    queries.append(text[ent["start"]: ent["end"]])
                else:
                    ent_label = random.choice(ent_labels)
                    queries.append(self.dictionary(ent_label)["description"])
                labels.append(ent_labels)

        all_candidates = []
        pbar = tqdm(total=(math.ceil(len(queries)/batch_size)), desc='Retrieve candidates')
        for i in range(0, len(queries), batch_size):
            pbar.update()
            batch_queries = queries[i:i + batch_size]
            batch_labels = labels[i:i + batch_size]
            _, batch_indices = self.retriever.search_knn(batch_queries, top_k=top_k, ignore_ids=batch_labels if only_negative else None)
            all_candidates.extend(batch_indices)
        pbar.close()

        return all_candidates



import logging
from dataclasses import dataclass
from typing import Any, Optional, Union

import torch
from transformers import BatchEncoding
from transformers.data.data_collator import pad_without_fast_tokenizer_warning

from entity_linkings.data_utils import (
    CollatorBase,
    EntityDictionary,
)

logger = logging.getLogger(__name__)


@dataclass
class CollatorForFEVRY(CollatorBase):
    dictionary: Optional[EntityDictionary] = None
    train: bool = False

    def __call__(self, features: list[Union[dict[str, Any], BatchEncoding]]) -> BatchEncoding:
        if self.dictionary is None:
            raise ValueError("this modules needs to dictionary.")

        features = [f.copy() for f in features]
        start_positions, end_positions = [], []
        batch_features, batch_candidates, batch_labels = [], [], []
        for f in features:
            new_features = {}
            new_features["input_ids"] = f.pop("input_ids")
            new_features["attention_mask"] = f.pop("attention_mask")
            token_type_ids = f.pop("token_type_ids", None)
            if token_type_ids is not None:
                new_features["token_type_ids"] = token_type_ids
            batch_features.append(new_features)

            start_position = f.pop("start_positions")
            end_position = f.pop("end_positions")
            start_positions.append(start_position)
            end_positions.append(end_position)

            golds = f.pop("labels", None)
            gold_ids = [self.dictionary.get_label_id(gold) for gold in golds] if golds is not None else []
            candidates = f.pop("candidates")
            candidate_ids = [self.dictionary.get_label_id(cand) for cand in candidates]

            if self.train:
                if not gold_ids:
                    raise ValueError("Gold labels are required during training.")
                cand_ids = [gold_ids[0]] + candidate_ids[:-1]
                batch_labels.append(0)
            else:
                cand_ids = candidate_ids
                if gold_ids:
                    gold_label_ids = [1 if cid in gold_ids else 0 for cid in cand_ids]
                    if 1 not in gold_label_ids:
                        batch_labels.append(-1)
                    else:
                        batch_labels.append(gold_label_ids.index(1))
            batch_candidates.append(cand_ids)

        batch = pad_without_fast_tokenizer_warning(
            self.tokenizer,
            batch_features,
            padding=self.padding,
            max_length=self.max_length,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors=self.return_tensors,
        )

        batch["start_positions"] = torch.tensor(start_positions) if self.return_tensors == "pt" else start_positions
        batch["end_positions"] = torch.tensor(end_positions) if self.return_tensors == "pt" else end_positions
        batch["candidates_ids"] = torch.tensor(batch_candidates) if self.return_tensors == "pt" else batch_candidates
        if batch_labels:
            batch["labels"] = torch.tensor(batch_labels) if self.return_tensors == "pt" else batch_labels

        return batch
